import pytest
from tests.metaschema.properties.test_MetaschemaProperty import (
    TestMetaschemaProperty as base_class)
import weakref


class DummyMeta(type):  # pragma: debug
    
    def __eq__(self, solf):
        import inspect
        a_str = f"{self.__module__}.{self.__name__}[{inspect.getfile(self)}]"
        b_str = f"{solf.__module__}.{solf.__name__}[{inspect.getfile(solf)}]"
        return a_str == b_str

    def __hash__(self):
        import inspect
        return hash(f"{self.__module__}.{self.__name__}[{inspect.getfile(self)}]")


class Dummy(metaclass=DummyMeta):  # pragma: debug
    pass


class ValidArgsClass1(metaclass=DummyMeta):
    test_args = tuple([int(0), Dummy])
    test_kwargs = dict(c=int(1), d=Dummy)
    valid_args = [{'type': 'int'}, {'type': 'class'}]
    valid_kwargs = {'c': {'type': 'int'},
                    'd': {'type': 'class'}}
    invalid_args = [{'type': 'int'}, {'type': 'float'}]
    invalid_kwargs = {'c': {'type': 'float'},
                      'd': {'type': 'float'}}

    def __init__(self, a, b, c=0, d=Dummy):
        self.b = b
        self.d = d
        self._input_args = tuple([a, weakref.ref(b)])
        self._input_kwargs = {'c': c, 'd': weakref.ref(d)}
        
    def __eq__(self, solf):
        if solf.__class__ != self.__class__:  # pragma: debug
            return False
        return ((self._input_args == solf._input_args)
                and (self._input_kwargs == solf._input_kwargs))


class ValidArgsClass2(ValidArgsClass1):
    def get_input_args(self):
        return self._input_args

    def get_input_kwargs(self):
        return self._input_kwargs
    

class ValidArgsClass3(ValidArgsClass2):
    @property
    def input_arguments(self):
        return self._input_args

    @property
    def input_keyword_arguments(self):
        return self._input_kwargs


class ValidArgsClass4(ValidArgsClass1):

    def __init__(self, *args, **kwargs):
        self._input_args = args
        self._input_kwargs = kwargs


class InvalidArgsClass:  # pragma: no cover
    # Old style class dosn't inherit from object
    pass


class TestArgsMetaschemaProperty(base_class):
    r"""Test class for ArgsMetaschemaProperty class."""
    
    _mod = 'yggdrasil.metaschema.properties.ArgsMetaschemaProperty'
    _cls = 'ArgsMetaschemaProperty'

    @pytest.fixture(scope="class")
    def valid_classes(self):
        return [ValidArgsClass1, ValidArgsClass2,
                ValidArgsClass3, ValidArgsClass4]

    @pytest.fixture(scope="class")
    def valid_instances(self, valid_classes):
        return [cls(*(cls.test_args), **(cls.test_kwargs))
                for cls in valid_classes]
    
    @pytest.fixture(scope="class")
    def valid(self, valid_instances):
        r"""Objects that are valid."""
        return [(x, x.__class__.valid_args) for x in valid_instances]

    @pytest.fixture(scope="class")
    def invalid(self, valid_instances):
        r"""Objects that are invalid."""
        return [(x, x.__class__.invalid_args) for x in valid_instances]

    @pytest.fixture(scope="class")
    def valid_type(self, valid_classes):
        r"""Valid type."""
        return valid_classes[0].valid_args
    
    @pytest.fixture(scope="class")
    def invalid_type(self, valid_classes):
        r"""Invalid type."""
        return valid_classes[0].invalid_args

    @pytest.fixture(scope="class")
    def encode_errors(self):
        r"""Object that enduce errors during encoding."""
        return [int(1), InvalidArgsClass]

    @pytest.fixture(scope="class")
    def valid_compare(self, valid_type):
        r"""Objects that successfully compare."""
        return [(valid_type, valid_type)]

    @pytest.fixture(scope="class")
    def invalid_compare(self, valid_type, invalid_type):
        r"""Objects that do not successfully compare."""
        return [(valid_type, invalid_type)]



import numpy as np


temp = np.array([400.0, 410.0, 420.0, 430.0, 440.0, 450.0])  # Temp in K
theta_i = np.linspace(0.0, np.pi / 2, 10)  # radians
n1 = 1.0
n2 = 1.5


for i in range(len(temp)):
    in1 = n1 - 0.0001 * (400.0 - temp[i])
    in2 = n2 - 0.05 * (400.0 - temp[i])
    theta_t = np.arcsin(in1 * np.sin(theta_i) / in2)
    R = np.abs((in1 * np.cos(theta_i) - in2 * np.cos(theta_t))
               / (in1 * np.cos(theta_i) + in2 * np.cos(theta_t)))**2
    print("Reflectance at %s K: %s" % (temp[i], R))



import os
import uuid
import pytest
from yggdrasil.communication import open_file_comm


def test_open_file_comm():
    r"""Test file comm context manager."""
    fname = os.path.join(os.path.dirname(__file__),
                         '%s.txt' % str(uuid.uuid4()))
    try:
        msg = b'Test message'
        with open_file_comm(fname, 'w') as comm:
            flag = comm.send(msg)
            assert(flag)
        with open_file_comm(fname, 'a') as comm:
            flag = comm.send(msg)
            assert(flag)
        with open_file_comm(fname, 'r') as comm:
            flag, res = comm.recv()
            assert(flag)
        assert(res == 2 * msg)
    finally:
        if os.path.isfile(fname):
            os.remove(fname)
    with pytest.raises(ValueError):
        with open_file_comm(fname, 'invalid') as comm:  # pragma: debug
            pass


__all__ = []



import pytest
import os
from tests.examples import TestExample as base_class


@pytest.mark.extra_example
class TestExampleRPC1(base_class):
    r"""Test the rpc_lesson1 example."""

    niter = 3

    @pytest.fixture(scope="class")
    def example_name(self):
        r"""str: Name of example being tested."""
        return "rpc_lesson1"

    @pytest.fixture
    def results(self):
        r"""Result that should be found in output files."""
        prev1 = 0
        prev2 = 1
        res = ''
        for i in range(self.niter):
            next = prev1 + prev2
            res += 'fib(%-2d) = %-2d\n' % ((i + 1), next)
            prev2 = prev1
            prev1 = next
        return [res]

    @pytest.fixture
    def output_files(self, tempdir):
        r"""Output file."""
        return [os.path.join(tempdir, 'client_output.txt')]



import argparse
import numpy as np
import os

def calculate_concentration(doy, dist, height, offset=60.0):
    r"""Function that calculates the concentration of CO2.

    Args:
        doy (float): Day of year.
        dist (float): Distance from the plant in cm.
        height (float): Distance from the ground in cm.
        offset (float, optional): Offset in the year in days. Defaults to 60.

    Returns:
        float: CO2 concentration in cm^-3

    """
    return np.sin(2.0 * np.pi * (doy + offset) / 365) / (dist * dist * height)


# Parse command-line arguments
parser = argparse.ArgumentParser("Calculate the co2 concentration at a given distance from a plant.")
parser.add_argument('dist', help='Distance from the plant (cm)', type=float)
parser.add_argument('height', help='Distance from the ground (cm)', type=float)
parser.add_argument('doy', help='Day of year', type=float)
args = parser.parse_args()
dist = args.dist
height = args.height
doy = args.doy
offset = 60.0

# Check if model is running as a part of an yggdrasil integration
with_yggdrasil = os.environ.get('YGG_SUBPROCESS', False)

# If the model is running as part of an yggdrasil integration, import
# the relevant yggdrasil routines and use the interface routine to
# complete the connection defined in the YAML
if with_yggdrasil:
    from yggdrasil import units
    from yggdrasil.languages.Python.YggInterface import YggInput, YggOutput
    height_in = YggInput('height')
    conc_out = YggOutput('co2')

    # Add units to parameters
    dist = units.add_units(dist, 'cm')
    offset = units.add_units(offset, 'days')

    # Loop over input
    while True:
        # Receive height
        flag, height_data = height_in.recv()
        if not flag:
            print('End of height input')
            break
        [doy, height] = height_data[:]

        
        # Compute concentration
        conc = calculate_concentration(doy, dist, height, offset=offset)
        print('Concentration', conc)

        # Send output
        flag = conc_out.send(doy, conc)
        if not flag:
            raise Exception("Error sending concentration to output")

else:
    
    # Compute concentration
    conc = calculate_concentration(doy, dist, height)
    print('Concentration', conc)



import pytest
from tests.serialize import TestSerializeBase as base_class


class TestPickleSerialize(base_class):
    r"""Test class for TestPickleSerialize class."""

    @pytest.fixture(scope="class", autouse=True)
    def serializer(self):
        r"""str: Serializer being tested."""
        return "pickle"

    def test_get_first_frame(self, python_class):
        r"""Test get_first_frame for empty message."""
        assert(python_class.get_first_frame(b'not a pickle') == b'')



import numpy as np
from yggdrasil import units
from yggdrasil.languages.Python.YggInterface import YggInput, YggOutput


def calc(doy, height, co2):
    return 3.0e8 * np.cos(2.0 * np.pi * doy / units.add_units(365, 'days')) / (co2 * height**3)


# Create input and output comms
height_in = YggInput('height')
conc_in = YggInput('co2')
humidity_out = YggOutput('humidity')

# Keep looping until there are no more inputs
while True:

    # Receive height
    flag, height_data = height_in.recv()
    if not flag:
        print("No more inputs")
        break
    [doy_height, height] = height_data[:]

    # Receive co2
    flag, [doy_co2, co2] = conc_in.recv()
    if not flag:
        raise Exception("Error receiving co2 input")

    # Check that height and co2 days match
    assert(doy_height == doy_co2)
    doy = doy_height

    # Call calculation
    humidity = calc(doy, height, co2)
    print('Humidity', doy, height, co2, humidity)

    # Send humidity
    flag = humidity_out.send(doy, height, co2, humidity)
    if not flag:
        raise Exception("Error sending humidity to output")



import pytest
from tests.drivers.test_BuildModelDriver import (
    TestBuildModelDriver as base_class)
import os
from yggdrasil.drivers.MakeModelDriver import MakeModelDriver, MakeCompiler


@pytest.mark.related_language('make')
def test_MakeCompiler():
    r"""Test MakeCompiler class."""
    assert(MakeCompiler.get_output_file(None, target='clean') == 'clean')


@pytest.mark.absent_language('make')
def test_MakeModelDriver_no_C_library(scripts):  # pragma: windows
    r"""Test MakeModelDriver error when C library not installed."""
    with pytest.raises(RuntimeError):
        MakeModelDriver('test', scripts['make'])


@pytest.mark.language('make')
def test_MakeModelDriver_error_notarget(scripts):
    r"""Test MakeModelDriver error for invalid target."""
    makedir, target = os.path.split(scripts['make'])
    with pytest.raises(RuntimeError):
        MakeModelDriver('test', 'invalid', makedir=makedir)


@pytest.mark.language('make')
def test_MakeModelDriver_error_nofile(scripts):
    r"""Test MakeModelDriver error for missing Makefile."""
    makedir, target = os.path.split(scripts['make'])
    with pytest.raises(RuntimeError):
        MakeModelDriver('test', 'invalid')


class TestMakeModelDriver(base_class):
    r"""Test runner for MakeModelDriver."""

    @pytest.fixture(scope="class")
    def language(self):
        r"""str: Language being tested."""
        return 'make'

    @pytest.fixture
    def makefile(self, sourcedir):
        r"""Makefile."""
        return os.path.join(sourcedir, 'Makefile')

    @pytest.fixture
    def instance_kwargs(self, testing_options, timeout, working_dir,
                        polling_interval, namespace, source, makefile):
        r"""Keyword arguments for a new instance of the tested class."""
        return dict(testing_options.get('kwargs', {}),
                    yml={'working_dir': working_dir},
                    timeout=timeout, sleeptime=polling_interval,
                    namespace=namespace, makefile=makefile)


class TestMakeModelDriver_wd(TestMakeModelDriver):
    r"""Test runner for MakeModelDriver with working directory."""

    @pytest.fixture
    def instance_kwargs(self, testing_options, timeout, sourcedir,
                        polling_interval, namespace, source):
        r"""Keyword arguments for a new instance of the tested class."""
        return dict(testing_options.get('kwargs', {}),
                    yml={'working_dir': sourcedir},
                    timeout=timeout, sleeptime=polling_interval,
                    namespace=namespace)

    def test_compile_model(self, target, instance):
        r"""Test compile model with alternate set of input arguments."""
        src = [target + '.c']
        instance.compile_model(target=target)
        instance.compile_model(source_files=src)
        with pytest.raises(RuntimeError):
            instance.compile_model(source_files=src,
                                   target=target + 'invalid')
        

class TestMakeModelDriver_wd_rel(TestMakeModelDriver):
    r"""Test runner for MakeModelDriver with makedir rel to working_dir."""

    @pytest.fixture
    def instance_kwargs(self, testing_options, timeout, working_dir,
                        polling_interval, namespace, source, sourcedir):
        r"""Keyword arguments for a new instance of the tested class."""
        makedir_parts = os.path.split(sourcedir)
        return dict(testing_options.get('kwargs', {}),
                    yml={'working_dir': working_dir},
                    timeout=timeout, sleeptime=polling_interval,
                    namespace=namespace, working_dir=makedir_parts[0],
                    makedir=makedir_parts[1])



import pytest
import numpy as np
from yggdrasil import units
from tests.serialize import TestSerializeBase as base_class


_options = [
    {},
    {'table_example': True, 'include_oldkws': True},
    {'table_example': True, 'include_oldkws': True,
     'array_columns': True},
    {'explicit_testing_options': {
        'kwargs': {'datatype':
                   {'type': 'array',
                    'items': {'type': '1darray',
                              'subtype': 'float',
                              'precision': 64}}},
        'empty': [],
        'objects': [[np.zeros(3, 'float'), np.zeros(3, 'float')],
                    [np.ones(3, 'float'), np.ones(3, 'float')]],
        'extra_kwargs': {},
        'typedef': {'type': 'array', 'items': {'type': '1darray',
                                               'subtype': 'float',
                                               'precision': 64}},
        'dtype': np.dtype("float64")}},
    {'explicit_testing_options': {
        'kwargs': {'datatype':
                   {'type': 'array',
                    'items': {'type': '1darray',
                              'subtype': 'float',
                              'precision': 64}},
                   'field_names': [b'a', b'b'],
                   'field_units': [b'cm', b'g']},
        'empty': [],
        'objects': [[units.add_units(np.zeros(3, 'float'), 'cm'),
                     units.add_units(np.zeros(3, 'float'), 'g')],
                    [units.add_units(np.ones(3, 'float'), 'cm'),
                     units.add_units(np.ones(3, 'float'), 'g')]],
        'extra_kwargs': {},
        'typedef': {'type': 'array',
                    'items': [{'type': '1darray',
                               'subtype': 'float',
                               'precision': 64,
                               'title': 'a',
                               'units': 'cm'},
                              {'type': '1darray',
                               'subtype': 'float',
                               'precision': 64,
                               'title': 'b',
                               'units': 'g'}]},
        'dtype': np.dtype([('a', '<f8'), ('b', '<f8')]),
        'field_names': ['a', 'b'],
        'field_units': ['cm', 'g']}},
    {'explicit_testing_options': {
        'kwargs': {'datatype': {'type': 'float'}},
        'empty': b'',
        'objects': [float(x) for x in range(5)],
        'extra_kwargs': {},
        'typedef': {'type': 'float', 'precision': 64},
        'dtype': None}}]


class TestDefaultSerialize(base_class):
    r"""Test class for DefaultSerialize class."""

    @pytest.fixture(scope="class", autouse=True)
    def serializer(self):
        r"""str: Serializer being tested."""
        return "default"

    @pytest.fixture(scope="class", autouse=True, params=_options)
    def options(self, request):
        r"""Arguments that should be provided when getting testing options."""
        return request.param

    def test_serialize_no_format(self, instance_kwargs, class_name,
                                 testing_options, instance, map_sent2recv,
                                 header_info):
        r"""Test serialize/deserialize without format string."""
        if (len(instance_kwargs) == 0) and (class_name == 'DefaultSerialize'):
            for iobj in testing_options['objects']:
                msg = instance.serialize(iobj, header_kwargs=header_info)
                iout, ihead = instance.deserialize(msg)
                assert(iout == map_sent2recv(iobj))
                # assert(ihead == header_info)
            # with pytest.raises(Exception):
            #     instance.serialize(['msg', 0])

    def test_invalid_update(self, instance_kwargs, class_name,
                            instance):
        r"""Test error raised when serializer updated with type that isn't
        compatible."""
        if (len(instance_kwargs) == 0) and (class_name == 'DefaultSerialize'):
            instance.initialize_from_message(np.int64(1))
            with pytest.raises(RuntimeError):
                instance.update_serializer(datatype={'type': 'ply'})



def get_testing_options():
    r"""Get testing parameters for this example."""
    return dict(
        input_files='input.txt',
        temp_output_dir=True,
        output_files=['client_output.txt'])



import os
import pickle
import numpy as np
from yggdrasil import tools, units
from yggdrasil.runner import run
import matplotlib.pyplot as plt
tyk_dir = 'tyk6'

if __name__ == '__main__':
    try:
        # Change to the solution directory
        old_dir = os.getcwd()
        if not old_dir.endswith(tyk_dir):
            os.chdir(os.path.join('solutions', tyk_dir))
        if not os.path.isdir('output'):
            os.mkdir('output')

        # Part 1: Integrate microbe w/ root model
        tools.display_source_diff('../../models/microbe.py',
                                  'models/microbe.py', number_lines=True)
        tools.display_source('yamls/microbe.yml', number_lines=True)
        run(['yamls/roots_v1.yml', 'yamls/microbe.yml', 'yamls/timesync.yml'],
            production_run=True)

        # Part 2: Integrate microbe w/ root & shoot model
        run(['yamls/roots_v1.yml', 'yamls/microbe.yml', 'yamls/timesync.yml',
             'yamls/shoot_v3.yml', 'yamls/light_v1_python.yml'],
            production_run=True)

        # Part 3: Plot mass as a function of time
        filename_root_masses = 'output/masses.pkl'
        with open(filename_root_masses, 'rb') as fd:
            root_masses = pickle.load(fd)
            root_masses['times'] = units.add_units(
                np.array(root_masses['times']), 'days')
            root_masses['masses'] = units.add_units(
                np.array(root_masses['masses']), 'kg')
            root_masses['times'].convert_to_mks()
            root_masses['masses'].convert_to_mks()
        filename_microbe_masses = 'output/microbe_masses.pkl'
        with open(filename_microbe_masses, 'rb') as fd:
            microbe_masses = pickle.load(fd)
            microbe_masses['times'] = units.add_units(
                np.array(microbe_masses['times']), 'min')
            microbe_masses['masses'] = units.add_units(
                np.array(microbe_masses['masses']), 'g')
            microbe_masses['times'].convert_to_mks()
            microbe_masses['masses'].convert_to_mks()
        plt.plot(root_masses['times'], root_masses['masses'])
        plt.plot(microbe_masses['times'], microbe_masses['masses'])
        plt.show()

    finally:
        os.chdir(old_dir)



import pytest
from tests.communication.test_CommBase import TestComm as base_class


class TestForkComm(base_class):
    r"""Tests for ForkComm communication class."""

    test_error_send = None
    test_error_recv = None
    test_work_comm = None
    test_send_recv_raw = None

    @pytest.fixture(scope="class", autouse=True)
    def python_class(self):
        r"""Python class that is being tested."""
        from yggdrasil.communication.ForkComm import ForkComm
        return ForkComm
    
    @pytest.fixture(scope="class", autouse=True)
    def commtype(self):
        r"""Communicator type being tested."""
        return "fork"

    @pytest.fixture(scope="class", autouse=True)
    def ncomm(self):
        r"""Number of communicators to include in the fork."""
        return 2

    @pytest.fixture(scope="class", autouse=True,
                    params=['broadcast', 'cycle', 'scatter'])
    def send_pattern(self, request):
        r"""Pattern in which to send messages to fork communicators."""
        return request.param

    @pytest.fixture(scope="class", autouse=True)
    def recv_pattern(self, send_pattern):
        r"""Pattern in which to recv messages to fork communicators."""
        pattern_map = {'broadcast': 'cycle',
                       'cycle': 'cycle',
                       'scatter': 'gather'}
        return pattern_map[send_pattern]

    @pytest.fixture(scope="class", autouse=True)
    def duplicate_msg(self, ncomm, send_pattern, recv_pattern):
        def wrapped_duplicate_msg(out, direction='send'):
            r"""Copy a message for 'scatter' communication pattern."""
            if ((((direction == 'send') and (send_pattern == 'scatter'))
                 or ((direction == 'recv') and (recv_pattern == 'gather')))):
                out = [out for _ in range(ncomm)]
            return out
        return wrapped_duplicate_msg

    @pytest.fixture(scope="class")
    def testing_options(self, python_class, options, ncomm, send_pattern,
                        recv_pattern, duplicate_msg):
        r"""Testing options."""
        out = python_class.get_testing_options(**options)
        out['kwargs'].update(ncomm=ncomm, pattern=send_pattern,
                             commtype='ForkComm')
        out.setdefault('recv_kwargs', {})
        out['recv_kwargs'].update(pattern=recv_pattern,
                                  commtype='ForkComm')
        out['msg'] = duplicate_msg(out['msg'])
        return out
    
    @pytest.fixture(scope="class")
    def process_send_message(self, duplicate_msg):
        r"""Factory for method to finalize messages for sending."""
        def wrapped_process_send_message(obj):
            return duplicate_msg(obj)
        return wrapped_process_send_message
    
    @pytest.fixture(scope="class")
    def map_sent2recv(self, ncomm, send_pattern, recv_pattern):
        r"""Factory for method to convert sent messages to received."""
        def wrapped_map_sent2recv(obj):
            r"""Convert a sent object into a received one."""
            if (((send_pattern == 'scatter')
                 and isinstance(obj, list) and obj)):
                single_obj = obj[0]
            else:
                single_obj = obj
            if recv_pattern == 'gather':
                if obj in [b'', []]:
                    return []
                return [single_obj for _ in range(ncomm)]
            return single_obj
        return wrapped_map_sent2recv

    @pytest.fixture(scope="class")
    def n_msg_expected(self, ncomm, send_pattern, recv_pattern):
        r"""Number of expected messages."""
        if (((send_pattern in ['broadcast', 'scatter'])
             and (recv_pattern == 'cycle'))):
            return ncomm
        return 1
    
    def test_send_recv_eof_no_close(self, send_comm, recv_comm, do_send_recv):
        r"""Test send/recv of EOF message with no close."""
        recv_comm.close_on_eof_recv = False
        for x in recv_comm.comm_list:
            x.close_on_eof_recv = False
        do_send_recv(send_comm, recv_comm,
                     send_params={'method': 'send_eof'})

    def test_send_recv_filter_eof(self, run_once, filtered_comms, send_comm,
                                  recv_comm, do_send_recv, timeout):
        r"""Test send/recv of EOF with filter."""
        do_send_recv(send_comm, recv_comm,
                     send_params={'method': 'send_eof'},
                     recv_params={'flag': False,
                                  'kwargs': {'timeout': 2 * timeout}})
        assert(recv_comm.is_closed)
        
    def test_send_recv_filter_recv_filter(self, filtered_comms,
                                          msg_filter_recv, send_comm,
                                          recv_comm, polling_interval,
                                          do_send_recv):
        r"""Test send/recv with filter that blocks recv."""
        do_send_recv(send_comm, recv_comm, msg_filter_recv,
                     recv_params={'message': recv_comm.empty_obj_recv,
                                  # Don't need to skip since
                                  # filter is evaluated after
                                  # receipt for fork comm
                                  'skip_wait': False,
                                  'count': 1,
                                  'kwargs': {'timeout': 10 * polling_interval}})
        
    def test_send_recv_after_close(self, commtype, send_comm, recv_comm,
                                   testing_options, ncomm):
        r"""Test that opening twice dosn't cause errors and that send/recv
        after close returns false."""
        send_comm.open()
        recv_comm.open()
        if 'rmq' in commtype:
            send_comm.bind()
            recv_comm.bind()
        send_comm.close()
        recv_comm.close()
        assert(send_comm.is_closed)
        assert(recv_comm.is_closed)
        flag = send_comm.send([testing_options['msg'] for _ in range(ncomm)])
        assert(not flag)
        flag, msg_recv = recv_comm.recv()
        assert(not flag)
        
    def test_async_gather(self, testing_options, send_pattern, recv_pattern,
                          send_comm, recv_comm, map_sent2recv, timeout):
        r"""Test scatter-gather w/ intermittent send."""
        if (send_pattern, recv_pattern) != ('scatter', 'gather'):
            pytest.skip("Only valid for scatter/gather pattern")
        test_msg = testing_options['msg']
        send_comm.comm_list[0].send(test_msg[0])
        flag, msg_recv = recv_comm.recv()
        assert(flag)
        assert(recv_comm.is_empty_recv(msg_recv))
        for msg_send, comm in zip(test_msg[1:], send_comm.comm_list[1:]):
            assert(comm.send(msg_send))
        flag, msg_recv = recv_comm.recv(timeout=timeout)
        assert(flag)
        assert(msg_recv == map_sent2recv(test_msg))


class TestForkCommList(TestForkComm):
    r"""Tests for ForkComm communication class with construction from address."""

    @pytest.fixture(scope="class", autouse=True)
    def send_pattern(self):
        r"""Pattern in which to send messages to fork communicators."""
        return 'broadcast'

    @pytest.fixture(scope="class")
    def testing_options(self, python_class, options, ncomm, send_pattern,
                        recv_pattern, duplicate_msg):
        r"""Testing options."""
        out = python_class.get_testing_options(**options)
        out['kwargs'].update(ncomm=ncomm, pattern=send_pattern,
                             commtype='ForkComm')
        out.setdefault('recv_kwargs', {})
        out['recv_kwargs'].update(
            pattern=recv_pattern, commtype='ForkComm',
            # To force test of construction from addresses
            comm_list=None)
        out['msg'] = duplicate_msg(out['msg'])
        return out



import pytest
from tests.communication.test_CommBase import TestComm as base_class


class TestRESTComm(base_class):
    r"""Test for RESTComm communication class."""

    @pytest.fixture(scope="class", autouse=True)
    def commtype(self):
        r"""Communicator type being tested."""
        return "rest"

    @pytest.fixture(scope="class", autouse=True)
    def running_service(self, running_service):
        with running_service('flask', partial_commtype='rest') as cli:
            yield cli



import pytest
import os
import numpy as np
import flaky
import copy
from yggdrasil.communication import get_comm
from yggdrasil.interface import YggInterface
from yggdrasil import constants
from yggdrasil.tools import get_YGG_MSG_MAX, is_lang_installed
from yggdrasil.components import import_component
from yggdrasil.drivers import ConnectionDriver
from tests import TestClassBase as base_class


YGG_MSG_MAX = get_YGG_MSG_MAX()


class ModelEnv(object):
    
    def __init__(self, language=None, **new_kw):
        new_kw['YGG_SUBPROCESS'] = 'True'
        if language is not None:
            new_kw['YGG_MODEL_LANGUAGE'] = language
        # Send environment keyword to fake language
        self.old_kw = {}
        for k, v in new_kw.items():
            self.old_kw[k] = os.environ.get(k, None)
            os.environ[k] = v
            
    def __enter__(self):
        return None

    def __exit__(self, type, value, traceback):
        for k, v in self.old_kw.items():
            if v is None:
                del os.environ[k]
            else:  # pragma: no cover
                os.environ[k] = v


def test_maxMsgSize():
    r"""Test max message size."""
    assert(YggInterface.maxMsgSize() == YGG_MSG_MAX)


def test_eof_msg():
    r"""Test eof message signal."""
    assert(YggInterface.eof_msg() == constants.YGG_MSG_EOF)


def test_bufMsgSize():
    r"""Test buf message size."""
    assert(YggInterface.bufMsgSize() == constants.YGG_MSG_BUF)


def test_init():
    r"""Test error on init."""
    with pytest.raises(Exception):
        YggInterface.YggInput('error')
    with pytest.raises(Exception):
        YggInterface.YggOutput('error')


def do_send_recv(language='python', fmt='%f\\n%d', msg=[float(1.0), np.int32(2)],
                 input_interface='YggInput', output_interface='YggOutput'):
    r"""Function to perform simple send/receive between two comms using a
    language interface that calls the Python interface.
    
    Args:
        language (str, optional): Language that should be mimicked for the
            interface test. Defaults to 'python'.
        fmt (str, optional): Format string to use for the test. Defaults to
            '%f\\n%d'.
        msg (object, optional): Message object that should be used for the
            test. Defaults to [float(1.0), int(2)].
        input_interface (str, optional): Name of the interface function/class
            that should be used for the test. Defaults to 'YggInput'.
        output_interface (str, optional): Name of the interface function/class
            that should be used for the test. Defaults to 'YggOutput'.

    """
    name = 'test_%s' % language
    # Set converter based on language driver
    ldrv = import_component('model', language)
    converter = ldrv.python2language
    # Create and start drivers to transport messages
    iodrv = ConnectionDriver.ConnectionDriver(
        name,
        inputs=[{'partner_model': 'model1', 'allow_multiple_comms': True}],
        outputs=[{'partner_model': 'model2', 'allow_multiple_comms': True}])
    iodrv.start()
    os.environ.update(iodrv.icomm.opp_comms)
    os.environ.update(iodrv.ocomm.opp_comms)
    # Connect and utilize interface under disguise as target language
    try:
        with ModelEnv(language=language, YGG_THREADING='True'):
            # Ensure start-up by waiting for signon message
            i = YggInterface.YggInit(input_interface, (name, fmt))
            i.drain_server_signon_messages()
            # Output
            o = YggInterface.YggInit(output_interface, (name, fmt))
            o.send(*msg)
            o.send_eof()
            o.close(linger=True)
            # Input
            assert(i.recv() == (True, converter(msg)))
            assert(i.recv() == (False, converter(constants.YGG_MSG_EOF)))
    finally:
        iodrv.terminate()


def test_YggInit_language():
    r"""Test access to YggInit via languages that call the Python interface."""
    for language in ['matlab', 'R']:
        if not is_lang_installed(language):
            continue
        do_send_recv(language=language)


def test_YggInit_backwards():
    r"""Check access to old class names for backwards compat."""
    do_send_recv(input_interface='CisInput',
                 output_interface='PsiOutput')


def test_YggInit_variables():
    r"""Test Matlab interface for variables."""
    assert(YggInterface.YggInit('YGG_MSG_MAX') == YGG_MSG_MAX)
    assert(YggInterface.YggInit('YGG_MSG_EOF') == constants.YGG_MSG_EOF)
    assert(YggInterface.YggInit('YGG_MSG_EOF')
           == YggInterface.YggInit('CIS_MSG_EOF'))
    assert(YggInterface.YggInit('YGG_MSG_EOF')
           == YggInterface.YggInit('PSI_MSG_EOF'))


class TestYggClass(base_class):
    r"""Test basic input/output to/from python/matlab."""

    _mod = 'yggdrasil.interface.YggInterface'
    fmt_str = '%5s\t%d\t%f\n'
    fmt_str_matlab = '%5s\\t%d\\t%f\\n'
    
    @pytest.fixture(scope="class", autouse=True,
                    params=['YggInput', 'YggOutput',
                            'YggAsciiFileInput', 'YggAsciiFileOutput',
                            'YggAsciiTableInput', 'YggAsciiTableOutput',
                            'YggAsciiArrayInput', 'YggAsciiArrayOutput',
                            'YggPickleInput', 'YggPickleOutput',
                            'YggPandasInput', 'YggPandasOutput',
                            'YggPlyInput', 'YggPlyOutput',
                            'YggObjInput', 'YggObjOutput'])
    def class_name(self, request):
        r"""Name of class that will be tested."""
        return request.param

    @pytest.fixture(scope="class", autouse=True, params=[None, 'matlab'])
    def interface_language(self, request, filecomm, direction):
        r"""str: Language being tested."""
        if ((request.param and filecomm
             and not ((filecomm == 'AsciiTableComm')
                      and (direction == 'output')))):
            pytest.skip("Redundent testing files for python and matlab.")
        return request.param

    @pytest.fixture(scope="class")
    def direction(self, class_name):
        r"""str: Direction of comm being tested."""
        if 'Input' in class_name:
            return 'input'
        else:
            return 'output'
        
    @pytest.fixture(scope="class")
    def filecomm(self, class_name):
        r"""str: File communicator to test."""
        if class_name in ['YggAsciiFileInput', 'YggAsciiFileOutput']:
            return 'AsciiFileComm'
        elif class_name in ['YggAsciiTableInput', 'YggAsciiTableOutput',
                            'YggAsciiArrayInput', 'YggAsciiArrayOutput']:
            return 'AsciiTableComm'
        elif class_name in ['YggPickleInput', 'YggPickleOutput']:
            return 'PickleFileComm'
        elif class_name in ['YggPandasInput', 'YggPandasOutput']:
            return 'PandasFileComm'
        elif class_name in ['YggPlyInput', 'YggPlyOutput']:
            return 'PlyFileComm'
        elif class_name in ['YggObjInput', 'YggObjOutput']:
            return 'ObjFileComm'
        return None

    @pytest.fixture
    def instance_args(self, name, class_name, format_str):
        r"""Arguments for a new instance of the tested class."""
        if class_name in ['YggAsciiTableOutput', 'YggAsciiArrayOutput']:
            return (name, format_str)
        return (name, )

    @pytest.fixture
    def instance_kwargs(self, testing_options, class_name, format_str):
        r"""Keyword arguments for a new instance of the tested class."""
        if class_name in ['YggInput', 'YggOutput']:
            return {'format_str': format_str}
        elif class_name == 'YggAsciiTableOutput':
            return {}
        return dict(testing_options.get('kwargs', {}))

    @pytest.fixture(scope="class")
    def format_str(self, interface_language):
        if interface_language == 'matlab':
            return self.fmt_str_matlab
        else:
            return self.fmt_str
    
    @pytest.fixture
    def name(self, uuid):
        return f"test{uuid.split('-')[0]}"

    @pytest.fixture
    def test_comm_kwargs(self):
        r"""dict: Keyword arguments for the test communicator."""
        return {}

    @pytest.fixture(scope="class")
    def filename(self):
        return os.path.join(os.getcwd(), 'temp_ascii.txt')

    @pytest.fixture(scope="class")
    def model1(self):
        r"""str: Name of one test model."""
        return 'model1'

    @pytest.fixture(scope="class")
    def model2(self):
        r"""str: Name of other test model."""
        return 'model2'

    @pytest.fixture(scope="class")
    def comm_class(self, filecomm):
        r"""Communicator class being tested."""
        if filecomm is None:
            return import_component('comm', 'default')
        else:
            return import_component('file', filecomm)

    @pytest.fixture(scope="class")
    def iodriver_class(self, direction, filecomm):
        r"""class: Input/output driver class."""
        if filecomm:
            return import_component('connection', 'file_' + direction)
        return ConnectionDriver.ConnectionDriver
        
    @pytest.fixture
    def iodriver_args(self, testing_options, name, model1, model2,
                      filecomm, filename, direction):
        r"""list: Connection driver arguments."""
        args = [name]
        kwargs = {'inputs': [{'partner_model': model1}],
                  'outputs': [{'partner_model': model2}]}
        if filecomm:
            args += [filename]
            if (direction == 'output'):
                filecomm_kwargs = copy.deepcopy(testing_options['kwargs'])
                filecomm_kwargs['filetype'] = filecomm
                return ([name, filename],
                        {'inputs': kwargs['inputs'],
                         'outputs': [filecomm_kwargs]})
            elif (direction == 'input'):
                filecomm_kwargs = copy.deepcopy(testing_options['kwargs'])
                filecomm_kwargs['filetype'] = filecomm
                return ([name, filename],
                        {'inputs': [filecomm_kwargs],
                         'outputs': kwargs['outputs']})
        return (args, kwargs)

    @pytest.fixture(scope="class", autouse=True)
    def options(self, class_name):
        r"""Arguments that should be provided when getting testing options."""
        if class_name in ['YggInput', 'YggOutput']:
            return {'table_example': True}
        elif class_name in ['YggAsciiArrayInput', 'YggAsciiArrayOutput']:
            return {'array_columns': True}
        # elif class_name in ['YggPandasInput', 'YggPandasOutput']:
        #     return {'as_frames': True}
        return {}

    @pytest.fixture(scope="class")
    def testing_options(self, comm_class, options):
        r"""Testing options."""
        return comm_class.get_testing_options(**options)

    @pytest.fixture(autouse=True)
    def input_file(self, filename, testing_options, direction, filecomm):
        r"""Create an input file."""
        if filecomm and (direction == 'input'):
            with open(filename, 'wb') as fd:
                fd.write(testing_options['contents'])
            assert(os.path.isfile(filename))
        try:
            yield
        finally:
            if filecomm and os.path.isfile(filename):
                os.remove(filename)

    @pytest.fixture(autouse=True)
    def iodriver(self, input_file, iodriver_class, iodriver_args,
                 verify_count_threads, verify_count_comms,
                 verify_count_fds):
        iodriver = iodriver_class(*iodriver_args[0], **iodriver_args[1])
        iodriver.start()
        iodriver.wait_for_loop()
        try:
            yield iodriver
        finally:
            iodriver.terminate()
            iodriver.cleanup()
            iodriver.disconnect()
            del iodriver

    @pytest.fixture(autouse=True)
    def test_comm(self, iodriver, direction, filecomm, test_comm_kwargs,
                  close_comm):
        r"""Communicator for testing."""
        test_comm = None
        if not filecomm:
            if direction == 'input':
                kws = iodriver.icomm.opp_comm_kwargs()
                kws.update(test_comm_kwargs)
                test_comm = get_comm('in', **kws)
            elif direction == 'output':
                kws = iodriver.ocomm.opp_comm_kwargs()
                kws.update(test_comm_kwargs)
                test_comm = get_comm('out', **kws)
        try:
            yield test_comm
        finally:
            if test_comm is not None:
                close_comm(test_comm)

    @pytest.fixture
    def model_env(self, direction, iodriver, model1, model2):
        r"""Environment variables that should be set for interface."""
        out = {}
        if direction == 'input':
            out.update(iodriver.ocomm.opp_comms,
                       YGG_MODEL_NAME=model2)
        elif direction == 'output':
            out.update(iodriver.icomm.opp_comms,
                       YGG_MODEL_NAME=model1)
        return out

    @pytest.fixture
    def instance(self, python_class, instance_args, instance_kwargs,
                 interface_language, model_env, close_comm):
        r"""New instance of the python class for testing."""
        with ModelEnv(language=interface_language, **model_env):
            out = python_class(*instance_args, **instance_kwargs)
            yield out
            out.is_interface = False
            close_comm(out)

    @pytest.fixture(autouse=True)
    def pandas_equality_patch(self, monkeypatch, pandas_equality):
        r"""Patch pandas DataFrame so that equals is used instead of '=='"""
        with monkeypatch.context() as m:
            import pandas
            m.setattr(pandas.DataFrame, '__eq__', pandas_equality)
            yield

    def test_msg(self, filecomm, testing_options, instance, timeout,
                 test_comm, iodriver, wait_on_function,
                 filename, direction, nested_approx, unyts_equality_patch):
        r"""Test sending/receiving message."""
        if direction == 'input':
            if filecomm:
                for msg_recv in testing_options['recv']:
                    msg_flag, msg_recv0 = instance.recv(timeout)
                    assert(msg_flag)
                    assert(msg_recv0 == nested_approx(msg_recv))
                msg_flag, msg_recv0 = instance.recv(timeout)
                assert(not msg_flag)
            else:
                for msg_send, msg_recv in zip(testing_options['send'],
                                              testing_options['recv']):
                    msg_flag = test_comm.send(msg_send)
                    assert(msg_flag)
                    msg_flag, msg_recv0 = instance.recv(timeout)
                    assert(msg_flag)
                    assert(msg_recv0 == nested_approx(msg_recv))
        else:
            if filecomm:
                for msg in testing_options['send']:
                    msg_flag = instance.send(msg)
                    assert(msg_flag)
                instance.send_eof()
                # Read temp file
                wait_on_function(lambda: not iodriver.ocomm.is_open)
                assert(os.path.isfile(filename))
                if testing_options.get('exact_contents', True):
                    with open(filename, 'rb') as fd:
                        res = fd.read()
                        assert(res == testing_options['contents'])
            else:
                for msg_send, msg_recv in zip(testing_options['send'],
                                              testing_options['recv']):
                    msg_flag = instance.send(msg_send)
                    assert(msg_flag)
                    msg_flag, msg_recv0 = test_comm.recv(timeout)
                    assert(msg_flag)
                    assert(msg_recv0 == nested_approx(msg_recv))


@flaky.flaky
class TestYggRpcClient(TestYggClass):
    r"""Test client-side RPC communication with Python."""

    @pytest.fixture(scope="class", autouse=True,
                    params=['YggRpcClient', 'YggRpcServer'])
    def class_name(self, request):
        r"""Name of class that will be tested."""
        return request.param

    @pytest.fixture(scope="class")
    def direction(self, class_name):
        r"""str: Direction of comm being tested."""
        if class_name == 'YggRpcClient':
            return 'output'
        else:
            return 'input'
    
    @pytest.fixture
    def instance_args(self, format_str, class_name, name, model1):
        r"""Arguments for a new instance of the tested class."""
        if class_name == 'YggRpcClient':
            return (f"{name}_{model1}", format_str, format_str)
        else:
            return (name, format_str, format_str)

    @pytest.fixture
    def test_comm_kwargs(self, class_name, format_str):
        r"""dict: Keyword arguments for the test communicator."""
        if class_name == 'YggRpcClient':
            return {'commtype': 'server',
                    'response_kwargs': {'format_str': format_str}}
        else:
            return {'commtype': 'client',
                    'response_kwargs': {'format_str': format_str}}
    
    @pytest.fixture(scope="class")
    def testing_options(self, comm_class, options):
        r"""Testing options."""
        out = comm_class.get_testing_options(**options)
        out.update(send=[[b'one', np.int32(1), 1.0]],
                   recv=[[b'one', np.int32(1), 1.0]])
        return out
    
    @pytest.fixture(autouse=True)
    def drain_signon(self, class_name, instance, test_comm):
        r"""Drain server signon messages."""
        if class_name == 'YggRpcClient':
            test_comm.drain_server_signon_messages()
        else:
            instance.drain_server_signon_messages()
        
    @pytest.fixture(scope="class")
    def iodriver_class(self):
        r"""class: Input/output driver class."""
        return import_component('connection', 'rpc_request')

    @pytest.fixture
    def iodriver_args(self, testing_options, name, model1, model2,
                      filecomm, filename):
        r"""list: Connection driver arguments."""
        args = [name]
        kwargs = {'inputs': [{'partner_model': model1,
                              'name': f"{model1}:{name}_{model1}"}],
                  'outputs': [{'partner_model': model2}]}
        return (args, kwargs)

    @pytest.fixture
    def model_env(self, direction, iodriver, model1, model2):
        r"""Environment variables that should be set for interface."""
        out = {}
        if direction == 'input':  # YggRpcServer
            out.update(iodriver.ocomm.opp_comms,
                       YGG_MODEL_NAME=model2,
                       YGG_NCLIENTS='1')
        elif direction == 'output':
            out.update(iodriver.icomm.opp_comms,
                       YGG_MODEL_NAME=model1)
        return out

    def test_msg(self, filecomm, testing_options, instance, timeout,
                 test_comm, iodriver, wait_on_function,
                 filename, direction, nested_approx, unyts_equality_patch):
        r"""Test sending/receiving message."""
        super(TestYggRpcClient, self).test_msg(
            filecomm, testing_options, instance, timeout,
            test_comm, iodriver, wait_on_function, filename, direction,
            nested_approx, unyts_equality_patch)
        if direction == 'output':
            send_comm = test_comm
            recv_comm = instance
        else:
            send_comm = instance
            recv_comm = test_comm
        for msg_send, msg_recv in zip(testing_options['send'],
                                      testing_options['recv']):
            msg_flag = send_comm.send(msg_send)
            assert(msg_flag)
            msg_flag, msg_recv0 = recv_comm.recv(timeout)
            assert(msg_flag)
            assert(msg_recv0 == nested_approx(msg_recv))



import pytest
import numpy as np
from yggdrasil import units
from tests.metaschema.properties.test_MetaschemaProperty import (
    TestMetaschemaProperty as base_class)


class TestLengthMetaschemaProperty(base_class):
    r"""Test class for LengthMetaschemaProperty class."""
    
    _mod = 'yggdrasil.metaschema.properties.ArrayMetaschemaProperties'
    _cls = 'LengthMetaschemaProperty'
    
    @pytest.fixture(scope="class")
    def nele(self):
        r"""Number of elements in test arrays."""
        return 3

    @pytest.fixture(scope="class")
    def valid_base(self, nele):
        r"""Base class for valid arrays."""
        return np.zeros(nele, 'float')

    @pytest.fixture(scope="class")
    def valid(self, nele, valid_base):
        r"""Objects that are valid."""
        return [(valid_base, nele), (units.add_units(valid_base, 'cm'), nele)]

    @pytest.fixture(scope="class")
    def invalid(self, nele, valid_base):
        r"""Objects that are invalid."""
        return [(valid_base, nele - 1)]

    @pytest.fixture(scope="class")
    def valid_compare(self, nele):
        r"""Objects that successfully compare."""
        return [(nele, nele)]

    @pytest.fixture(scope="class")
    def invalid_compare(self, nele):
        r"""Objects that do not successfully compare."""
        return [(nele - 1, nele), (nele, nele - 1)]


class TestShapeMetaschemaProperty(TestLengthMetaschemaProperty):
    r"""Test class for ShapeMetaschemaProperty class."""
    
    _mod = 'yggdrasil.metaschema.properties.ArrayMetaschemaProperties'
    _cls = 'ShapeMetaschemaProperty'
    
    @pytest.fixture(scope="class")
    def nele(self):
        r"""Number of elements in test arrays."""
        return (3, 4)

    @pytest.fixture(scope="class")
    def invalid(self, nele, valid_base):
        r"""Objects that are invalid."""
        return [(valid_base, (nele[0], nele[1] - 1))]

    @pytest.fixture(scope="class")
    def valid_compare(self, nele):
        r"""Objects that successfully compare."""
        return [(nele, nele), (nele, list(nele))]

    @pytest.fixture(scope="class")
    def invalid_compare(self, nele):
        r"""Objects that do not successfully compare."""
        return [(nele, nele[::-1]), (nele, nele[:-1])]



import pytest
from tests.serialize import TestSerializeBase as base_class


class TestAsciiMapSerialize(base_class):
    r"""Test class for TestAsciiMapSerialize class."""

    @pytest.fixture(scope="class", autouse=True)
    def serializer(self):
        r"""str: Serializer being tested."""
        return "map"

    def test_error_delim(self, instance):
        r"""Test error for message with too many delimiters on a line."""
        msg = instance.delimiter.join(
            ['args1', 'val1', 'args2', 'val2']).encode("utf-8")
        with pytest.raises(ValueError):
            instance.deserialize(msg)

    def test_error_nonstrval(self, instance):
        r"""Test error on serializing dictionary with non-string values."""
        obj = {1: 'here'}
        with pytest.raises(ValueError):
            instance.serialize(obj)

    def test_remove_quotes(self, instance):
        r"""Test deserialization of message with single quotes."""
        send_msg_list = [b"a\t'a_value'\n", b'a\t"a_value"\n',
                         b"a\ta_value\n"]
        recv_msg = {'a': 'a_value'}
        for send_msg in send_msg_list:
            assert(instance.deserialize(send_msg)[0] == recv_msg)



import os
import gc
import re
import sys
import glob
import copy
import shutil
import pytest
import logging
import argparse
import subprocess
import contextlib
from yggdrasil import platform, constants
from yggdrasil.tools import (
    get_supported_lang, get_supported_comm, get_supported_type)
from yggdrasil.components import import_component
from yggdrasil.multitasking import _on_mpi
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


_test_directory = os.path.dirname(__file__)
_test_registry = []
_weakref_registry = []
_markers = [
    ("long_running", "--long-running",
     "tests that take a long time to run", None),
    ("extra_example", "--extra-examples",
     "tests for superfluous examples", None),
    ("production_run", "--production-run", None)
]
_params = {
    "example_name": [],
    "language": sorted([x for x in constants.LANGUAGES['all']
                        if x not in ['mpi', 'dummy']]),
    "commtype": sorted(get_supported_comm()),
    "filetype": sorted(list(
        constants.COMPONENT_REGISTRY["file"]["subtypes"].keys())),
    "use_async": ['False', 'True'],
    "transform": None,
    "filter": None,
    "serializer": None,
    "typename": get_supported_type(),
}
_mpi_paths = [
    os.path.join("communication", "test_MPIComm.py"),
    os.path.join("examples", "test_gs_lesson4.py"),
    os.path.join("examples", "test_rpc_lesson3b.py"),
]
if platform._is_win:
    _example1_pattern = "[a-g]*"
else:
    _example1_pattern = "[A-Za-g]*"
    _mpi_paths.append(
        os.path.join("examples", "test_model_error_with_io.py"))
_suites = [
    ("top", "tests without an explicit suite", [""], []),
    ("examples", "examples", ["examples"], []),
    ("examples_part1", "1st half of examples",
     [os.path.join("examples", f"test_{_example1_pattern}*.py")], []),
    ("examples_part2", "2nd half of examples",
     [os.path.join("examples", "test_[h-z]*.py")], []),
    ("demos", "demonstrations", ["demos"], []),
    ("types", "type I/O in the supported languages",
     [os.path.join("examples", "test_types.py")], ['--long-running']),
    ("timing", "timing statistics tools", ["test_timing.py"],
     ["--long-running", "--production-run"]),
    ("comms", "communicators", ["communication"], []),
    ("files", "file communicators", ["communication"], []),
    ("connections", "connection drivers",
     [os.path.join("drivers", "test_ConnectionDriver.py"),
      os.path.join("drivers", "test_FileInputDriver.py"),
      os.path.join("drivers", "test_FileOutputDriver.py"),
      os.path.join("drivers", "test_RPCRequestDriver.py")], []),
    ("models", "model drivers",
     [os.path.join("drivers", "test_*ModelDriver.py")], []),
    ("mpi", "MPI based communication", _mpi_paths, []),
]


class DummyParser(argparse.ArgumentParser):

    def __init__(self, isolate_options=[]):
        self.isolate_options = isolate_options
        super(DummyParser, self).__init__()
        self.add_argument("file_or_dir", nargs="*")
        self.add_argument('-c', type=str, nargs=1)
        self.add_argument('-p', type=str, nargs=1)
        add_options_to_parser(self)

    def addoption(self, *args, **kwargs):
        name = args[0].lstrip('-').replace('-', '_')
        if self.isolate_options and not (
                (name in self.isolate_options)
                or (args[0] in self.isolate_options)):
            return
        self.add_argument(*args, **kwargs)

    def parse_known_and_unknown_args(self, *args, **kwargs):
        return self.parse_known_args(*args, **kwargs)


def setup_ci(args):
    import site
    top_dir = os.path.dirname(os.getcwd())
    for x in site.getsitepackages():
        package_dir = os.path.join(x, 'yggdrasil')
        if os.path.isdir(package_dir):
            break
    assert(os.path.isdir(package_dir))
    args += ['-v',
             '--import-mode=importlib',
             f'--cov={package_dir}',
             '-c', 'setup.cfg',
             '--cov-config=.coveragerc',
             '--ignore=yggdrasil/rapidjson/']
    # f'--rootdir={package_dir}']
    if not any(x.startswith('--with-mpi') for x in args):
        args += ['--reruns=2', '--reruns-delay=1', '--timeout=900']
    # Additional checks
    if not os.path.isfile('setup.cfg'):
        raise RuntimeError("The CI tests must be run from the root "
                           "directory of the yggdrasil git repository.")
    src_cmd = ('python -c \"import versioneer; '
               'print(versioneer.get_version())\"')
    dst_cmd = ('python -c \"import yggdrasil; '
               'print(yggdrasil.__version__)\"')
    src_ver = subprocess.check_output(src_cmd, shell=True)
    dst_ver = subprocess.check_output(dst_cmd, shell=True, cwd=top_dir)
    if src_ver != dst_ver:  # pragma: debug
        raise RuntimeError(("Versions do not match:\n"
                            "\tSource version: %s\n"
                            "\tBuild  version: %s\n")
                           % (src_ver, dst_ver))
    subprocess.check_call(
        ["flake8", "yggdrasil", "--append-config", "setup.cfg"])
    if os.environ.get("YGG_CONDA", None):
        subprocess.check_call(["python", "create_coveragerc.py"])
    if not os.path.isfile(".coveragerc"):
        raise RuntimeError(".coveragerc file dosn't exist.")
    with open(".coveragerc", "r") as fd:
        print(fd.read())
    subprocess.check_call(["yggdrasil", "info", "--verbose"])


def remove_option(args, options, remove_file_or_dir=False):
    if isinstance(options, str):
        options = [options]
    args_copy = copy.copy(args)
    # print("BEFORE", args)
    parser = DummyParser(isolate_options=options)
    parsed_args, remaining = parser.parse_known_and_unknown_args(args)
    if not remove_file_or_dir:
        remaining += parsed_args.file_or_dir
    args.clear()
    args += [x for x in args_copy if x in remaining]
    # print("AFTER ", args, set(args_copy) - set(args))


# def pytest_load_initial_conftests(args):
def pytest_cmdline_preparse(args, dont_exit=False):
    r"""Adjust the pytest arguments before testing."""
    # Check for run in separate process before adding CI args
    parser = DummyParser()
    pargs = parser.parse_known_and_unknown_args(args)[0]
    run_process = False
    prefix = []
    for k in ['separate_tests', 'suite']:
        if getattr(pargs, k, None) is None:
            setattr(pargs, k, [])
    # Disable output capture
    if pargs.nocapture:
        remove_option(args, 'nocapture')
        args += ['-s', '-o', 'log_cli=true']
    # MPI script
    if pargs.mpi_script:
        remove_option(args, 'mpi_script')
        mpi_test_args = [
            '--suite=mpi', f'--write-script={pargs.mpi_script}']
        if pargs.mpi_nproc > 1:
            remove_option(args, 'mpi_nproc')
            mpi_test_args.append(f'--mpi-nproc={pargs.mpi_nproc}')
            pargs.mpi_nproc = 1
        mpi_test_args = " ".join(mpi_test_args)
        args.append(f'--separate-test={mpi_test_args}')
        if not pargs.separate_tests:
            pargs.separate_tests = []
        pargs.separate_tests.append(mpi_test_args)
    # MPI process should be started
    if ('mpi' in pargs.suite) and (pargs.mpi_nproc <= 1) and (not _on_mpi):
        pargs.mpi_nproc = 2
    if pargs.mpi_nproc > 1:
        remove_option(args, 'mpi_nproc')
        run_process = True
        prefix = ['mpiexec', '-n', str(pargs.mpi_nproc)]
        if os.environ.get('CI', False) and platform._is_linux:
            prefix.append('--oversubscribe')
        if '--with-mpi' not in args:
            args.append('--with-mpi')
        args += ['-p', 'no:flaky']
        for x in ['--reruns=2', '--reruns-delay=1', '--timeout=900']:
            if x in args:
                args.remove(x)
    # Continuous integration
    if pargs.ci and (not _on_mpi):
        setup_ci(args)
        # remove_option(args, 'ci')
        args.remove('--ci')  # Much faster
        # Must launch in separate process so that pytest recognizes
        # the added --cov={install_dir} and --import-mode=importlib flags
        run_process = True
    # Write a script to call later
    if pargs.write_script:
        remove_option(args, 'write_script')
        if not os.path.isabs(pargs.write_script):
            pargs.write_script = os.path.abspath(pargs.write_script)
        write_pytest_script(pargs.write_script,
                            prefix + ['pytest'] + args)
        if dont_exit:
            return 0
        sys.exit(0)
    # Check for separate tests
    if pargs.separate_tests:
        remove_option(args, 'separate_tests')
    for x in pargs.separate_tests:
        x_args = x.split()
        x_args_copy = copy.copy(args)
        excluded = ([m[1] for m in _markers]
                    + ['suite', 'language', 'skip_language', 'default_comm']
                    + [f'parametrize_{k}' for k in _params.keys()])
        remove_option(x_args_copy, excluded, remove_file_or_dir=True)
        x_args += x_args_copy
        # x_args_keys = [k.split('=')[0] for k in x_args if k.startswith('-')]
        # for k in x_args_copy:
        #     if k.split('=')[0] not in x_args_keys:
        #         x_args.append(k)
        assert(any((k.split('=', 1)[0] == '--write-script') for k in x_args))
        if not pargs.second_attempt:
            pytest_cmdline_preparse(x_args, dont_exit=True)
    # Run test in separate process
    if run_process:
        flag = subprocess.call(prefix + ['pytest'] + args)
        if dont_exit:
            return flag
        sys.exit(flag)
    # Add test suites paths
    suite_map = {x[0]: (x[2], x[3]) for x in _suites}
    suite_files = []
    for suite in pargs.suite:
        for f in suite_map[suite][0]:
            suite_files += glob.glob(os.path.join(_test_directory, f))
        args += suite_map[suite][1]
    if suite_files:
        if not pargs.file_or_dir:
            args += ['--end-yggdrasil-opts'] + sorted(suite_files)
    elif not pargs.file_or_dir:
        args += ['--end-yggdrasil-opts'] + ["tests"]
    print(f"Updated args: {args}")


def pytest_addoption(parser):
    add_options_to_parser(parser)


def add_options_to_parser(parser):
    languages = sorted(get_supported_lang())
    for x in _markers:
        parser.addoption(x[1], action="store_true", default=False,
                         help=f"run {x[2]} tests")
    for k, v in _params.items():
        if v is None:
            v = sorted(list(constants.COMPONENT_REGISTRY[k]["subtypes"].keys()))
        choices = v if v else None
        parser.addoption(f"--parametrize-{k.replace('_', '-')}",
                         help=f"Set '{k}' test parameter", nargs='*',
                         choices=choices)
    parser.addoption('--second-attempt', action='store_true',
                     help=('Indicates the is the second attempt.'))
    parser.addoption('--ci', action='store_true',
                     help=('Perform additional operations required for '
                           'testing on continuous integration services.'))
    parser.addoption('--ygg-debug', action='store_true',
                     help=('Turn on debug level logging and increase the '
                           'scrutiny of messages during testing.'))
    parser.addoption('--ygg-loglevel', '--ygg-log-level', type=str,
                     help='Level of logging to use.')
    parser.addoption('--default-comm', type=str,
                     choices=sorted(
                         constants.COMPONENT_REGISTRY['comm']['subtypes'].keys()),
                     help="Communicator that should be used by default.")
    parser.addoption('--suite', '--suites', '--test-suite',
                     # type=str, action="extend",  # python >= 3.8
                     nargs='*', choices=[x[0] for x in _suites],
                     help="Test suite that should be run.")
    parser.addoption('--language', '--languages',
                     # type=str, action="extend",  # python >= 3.8
                     nargs='*', choices=languages,
                     help="Language(s) that should be tested.")
    parser.addoption('--skip-language', '--skip-languages',
                     # type=str, action="extend",  # python >= 3.8
                     nargs='*', choices=languages,
                     help="Language(s) that should be tested.")
    parser.addoption('--write-script', type=str,
                     help=("Name of script that should be created to run "
                           "tests."))
    parser.addoption('--mpi-nproc', type=int, default=1,
                     help="Number of MPI processes to run tests on.")
    parser.addoption('--mpi-script', type=str,
                     help=("Name of script that should be written to "
                           "run MPI tests. If --mpi-nproc is not set, it "
                           "will default to 2. If --mpi-nproc is set, it "
                           "will only be used in the MPI script that is "
                           "generated."))
    parser.addoption('--separate-tests', '--separate-test',
                     type=str, action="append",
                     help="Flags for an additional test that should be run")
    parser.addoption('--nocapture', action="store_true",
                     help="Don't capture output or log messages from tests.")
    parser.addoption('--end-yggdrasil-opts', action="store_true",
                     help="Internal use only")


def pytest_configure(config):
    for x in _markers:
        config.addinivalue_line("markers", f"{x[0]}: {x[2]}")
    config.addinivalue_line(
        "markers", ("suite(name,disabled=False,ignore=[]): mark test "
                    "as belonging to a group of tests, whether or not the "
                    "suite should be disabled by default, and suites "
                    "included by parent classes that should be ignored"))
    config.addinivalue_line(
        "markers", ("language(name): mark test as requiring a language is "
                    "installed"))
    config.addinivalue_line(
        "markers", ("absent_language(name): mark test as requiring a "
                    "language is NOT installed."))
    config.addinivalue_line(
        "markers", ("related_language(name): mark test as being related "
                    "to a language. The language may or may not be "
                    "installed, but must be enabled for the test to run."))


# def pytest_runtest_setup(item):
#     langs = [mark.args[0] for mark in item.iter_markers(name="language")]
#     if langs:
#         enabled = item.config.getoption("--language")
#         disabled = item.config.getoption("--skip-language")
#         if (((enabled and (not all(x in enabled for x in langs)))
#              or (disabled and any(x in disabled for x in langs)))):
#             pytest.skip(f"test requires languages {langs!r}")
#         elif not all(import_component('model', x).is_installed()
#                      for x in langs):
#             pytest.skip(f"test requires languages {langs!r}, "
#                         f"not all are installed")
#     absent_langs = [mark.args[0] for mark in
#                     item.iter_markers(name="absent_language")]
#     if absent_langs:
#         if any(import_component('model', x).is_installed()
#                for x in absent_langs):
#             pytest.skip(f"test requires languages {absent_langs!r} NOT "
#                         f"be installed")


def pytest_collection_modifyitems(config, items):
    for x in _markers:
        if config.getoption(x[1]):
            continue
        skip_x = pytest.mark.skip(reason=f"need {x[1]} to run")
        for item in items:
            if x[0] in item.keywords:
                item.add_marker(skip_x)
    # Handle suite & language markers
    selected_suites = config.getoption('--suite')
    enabled = config.getoption("--language")
    disabled = config.getoption("--skip-language")
    for item in items:
        # Suites
        suites = [mark.args[0] for mark in item.iter_markers(name="suite")]
        for mark in item.iter_markers(name="suite"):
            ignore = mark.kwargs.get('ignore', [])
            if isinstance(ignore, str):
                ignore = [ignore]
            for suite in ignore:
                while suite in suites:
                    suites.remove(suite)
        if 'examples' in suites:
            example_name = item.listnames()[-1].split('[')[-1].split('-')[0]
            if re.match(_example1_pattern, example_name):
                suites.append('examples_part1')
            else:
                suites.append('examples_part2')
        suites_disabled = [mark.kwargs.get('disabled', False)
                           for mark in item.iter_markers(name="suite")]
        if suites and not any(suites_disabled):
            suites.append('top')
        skip_x = None
        if selected_suites:
            if (((suites and not any(x in selected_suites for x in suites))
                 or (not suites and ('top' not in selected_suites)))):
                skip_x = pytest.mark.skip(
                    reason=f"none of test's suites ({suites}) selected.")
        elif any(suites_disabled):
            skip_x = pytest.mark.skip(
                reason=f"one of test's suites ({suites}) not selected.")
        if skip_x:
            item.add_marker(skip_x)
        # Selected/installed languages
        langs = [mark.args[0] for mark in item.iter_markers(name="language")]
        if langs:
            skip_x = None
            if (((enabled and (not all(x in enabled for x in langs)))
                 or (disabled and any(x in disabled for x in langs)))):
                skip_x = pytest.mark.skip(
                    reason=f"test requires languages {langs!r}")
            elif not all(import_component('model', x).is_installed()
                         for x in langs):
                skip_x = pytest.mark.skip(
                    reason=(f"test requires languages {langs!r}, "
                            f"not all are installed"))
            if skip_x:
                item.add_marker(skip_x)
        # Related languages
        related_langs = [mark.args[0] for mark in
                         item.iter_markers(name="related_language")]
        if related_langs:
            if (((enabled and (not all(x in enabled for x in related_langs)))
                 or (disabled and any(x in disabled for x in related_langs)))):
                item.add_marker(
                    pytest.mark.skip(
                        reason=(f"test is related to one or more languages "
                                f"{langs!r} that are not enabled or are "
                                f"disabled")))
        # Excluded languages
        absent_langs = [mark.args[0] for mark in
                        item.iter_markers(name="absent_language")]
        if absent_langs:
            if any(import_component('model', x).is_installed()
                   for x in absent_langs):
                item.add_marker(
                    pytest.mark.skip(
                        reason=(f"test requires languages {absent_langs!r} "
                                f"NOT be installed")))


def pytest_generate_tests(metafunc):
    for k, v in _params.items():
        if k not in metafunc.fixturenames:
            continue
        fixture = getattr(metafunc.cls, k, None)
        if ((fixture
             and ('request' not in
                  fixture.__wrapped__.__code__.co_varnames))):
            continue
        flag = f"--parametrize-{k.replace('_', '-')}"
        scope = None
        class_params = None
        if metafunc.cls and hasattr(metafunc.cls, f"parametrize_{k}"):
            class_params = getattr(metafunc.cls, f"parametrize_{k}")
            if callable(class_params):
                class_params = class_params(metafunc)
        if metafunc.config.getoption(flag):
            params = metafunc.config.getoption(flag)
            if k == 'use_async':
                params = [(x.lower() in ['true', '1']) for x in params]
            if isinstance(class_params, (list, tuple)):
                params = [x for x in params if x in class_params]
        elif isinstance(class_params, (list, tuple)):
            params = class_params
            scope = "class"
        else:
            if metafunc.cls:
                scope = "class"
            else:
                scope = None
            if v is None:
                params = sorted(list(
                    constants.COMPONENT_REGISTRY[k]["subtypes"].keys()))
            else:
                params = v
        metafunc.parametrize(k, params, indirect=True, scope=scope)


def write_pytest_script(fname, argv):
    r"""Write a script to run the pytest command.

    Args:
        fname (str): Full path to file where the script should be written.
        argv (list): Command options.

    """
    import stat
    from yggdrasil import platform
    cmd = ' '.join(argv)
    if platform._is_win:
        cmd = cmd.replace('\\', '/')
    if platform._is_win and (not os.environ.get("CONDA_PREFIX", None)):
        lines = [cmd + ' %*']
    else:
        lines = ['#!/bin/bash',
                 cmd + ' $@']
    with open(fname, 'w') as fd:
        fd.write('\n'.join(lines))
    os.chmod(fname, (stat.S_IRWXU
                     | stat.S_IRGRP | stat.S_IXGRP
                     | stat.S_IROTH | stat.S_IXOTH))
    contents = '\n\t' + '\n\t'.join(lines)
    print(f"Wrote test script to '{fname}':{contents}")


# Session level constants
@pytest.fixture(scope="session")
def project_dir():
    r"""Directory in which yggdrasil is installed."""
    import yggdrasil
    return os.path.abspath(os.path.dirname(yggdrasil.__file__))


@pytest.fixture(scope="session")
def logger():
    r"""Package logger."""
    import logging
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    return logger


@pytest.fixture(scope="session")
def MagicTestError():
    class MagicTestError(Exception):
        r"""Special exception for testing."""
        pass
    return MagicTestError


@pytest.fixture(scope="session")
def magic_error_replacement(MagicTestError):
    r"""Replacement for monkeypatching to raise an error."""
    def magic_error_replacement_w(*args, **kwargs):
        raise MagicTestError()
    return magic_error_replacement_w


@pytest.fixture(scope="session")
def timeout():
    r"""Time that should be waited during time outs."""
    return 10.0


@pytest.fixture(scope="session")
def polling_interval():
    r"""Time that should be waited between polls during time outs."""
    return 0.01


@pytest.fixture(scope="session")
def communicator_types():
    r"""list: Supported communicator types."""
    from yggdrasil.tools import get_supported_comm
    return get_supported_comm()


@pytest.fixture(scope="session")
def languages():
    r"""list: Supported languages."""
    from yggdrasil.tools import get_supported_lang
    return get_supported_lang()


@pytest.fixture(scope="session")
def scripts():
    r"""Dictionary of test scripts for each language."""
    from yggdrasil import tools
    script_dir = os.path.join(os.path.dirname(__file__), 'scripts')
    script_list = [
        ('c', ['gcc_model.c', 'hellofunc.c']),
        ('c++', ['gcc_model.cpp', 'hellofunc.c']),
        ('make', 'gcc_model'),
        ('cmake', 'gcc_model'),
        ('matlab', 'matlab_model.m'),
        ('matlab_error', 'matlab_error_model.m'),
        ('python', 'python_model.py'),
        ('error', 'error_model.py'),
        ('lpy', 'lpy_model.lpy'),
        ('r', 'r_model.R'),
        ('fortran', ['hellofunc.f90', 'fortran_model.f90']),
        ('sbml', 'sbml_model.xml'),
        ('osr', 'osr_model.xml')]
    scripts = {}
    for k, v in script_list:
        if isinstance(v, list):
            scripts[k] = [os.path.join(script_dir, iv) for iv in v]
        else:
            scripts[k] = os.path.join(script_dir, v)
    # scripts = {k: os.path.join(script_dir, v) for k, v in script_list}
    if platform._is_win and (not tools.in_powershell()):  # pragma: windows
        scripts['executable'] = ['ping', '-n', '1', '127.0.0.1']
    else:
        scripts['executable'] = ['sleep', 0.1]
    # Makefile
    if platform._is_win:  # pragma: windows
        makefile0 = os.path.join(script_dir, "Makefile_windows")
    else:
        makefile0 = os.path.join(script_dir, "Makefile_linux")
    dest = os.path.join(script_dir, "Makefile")
    shutil.copy(makefile0, dest)
    yield scripts
    if os.path.isfile(dest):
        os.remove(dest)


@pytest.fixture(scope="session")
def yamls():
    r"""Dictionary of test YAMLs for each language."""
    yaml_dir = os.path.join(os.path.dirname(__file__), 'yamls')
    yaml_list = [
        ('c', 'gcc_model.yml'),
        ('cpp', 'gpp_model.yml'),
        ('make', 'make_model.yml'),
        ('cmake', 'cmake_model.yml'),
        ('matlab', 'matlab_model.yml'),
        ('python', 'python_model.yml'),
        ('error', 'error_model.yml'),
        ('lpy', 'lpy_model.yml'),
        ('fortran', 'fortran_model.yml'),
        ('sbml', 'sbml_model.yml'),
        ('osr', 'osr_model.yml')]
    yamls = {k: os.path.join(yaml_dir, v) for k, v in yaml_list}
    return yamls


# Fixtures based on CLI options
@pytest.fixture(scope="session", autouse=True)
def config_env(pytestconfig):
    r"""Set environment variables based on CLI options."""
    second_attempt = pytestconfig.getoption("--second-attempt")
    production_run = pytestconfig.getoption("--production-run")
    default_comm = pytestconfig.getoption("--default-comm")
    debug = pytestconfig.getoption("--ygg-debug")
    loglevel = pytestconfig.getoption("--ygg-loglevel")
    if second_attempt:
        production_run = False
        debug = True
    from yggdrasil import config
    with config.temp_config(production_run=production_run,
                            debug=debug, default_comm=default_comm,
                            loglevel=loglevel, client_loglevel=loglevel):
        config.cfg_logging()
        yield
    config.cfg_logging()
    

# Fixtures for managing the test environment
@pytest.fixture(scope="session", autouse=True)
def utf8_encoding():
    r"""Set the encoding to utf-8 if it is not already."""
    lang = os.environ.get('LANG', '')
    if 'UTF-8' not in lang:  # pragma: debug
        os.environ['LANG'] = 'en_US.UTF-8'
    yield
    if 'UTF-8' not in lang:  # pragma: debug
        os.environ['LANG'] = lang


@pytest.fixture(scope="session")
def debug_log():
    r"""Set the log level to debug."""
    from yggdrasil.config import ygg_cfg, cfg_logging
    loglevel = ygg_cfg.get('debug', 'ygg')
    ygg_cfg.set('debug', 'ygg', 'DEBUG')
    cfg_logging()
    yield
    if loglevel is not None:
        ygg_cfg.set('debug', 'ygg', loglevel)
        cfg_logging()


@pytest.fixture(scope="session")
def change_default_comm():
    r"""Set the default comm."""
    @contextlib.contextmanager
    def change_default_comm_w(default_comm):
        from yggdrasil.communication.DefaultComm import DefaultComm
        old_default_comm = os.environ.get('YGG_DEFAULT_COMM', None)
        if default_comm is None:
            os.environ.pop('YGG_DEFAULT_COMM', None)
        else:
            os.environ['YGG_DEFAULT_COMM'] = default_comm
        DefaultComm._reset_alias()
        yield
        del os.environ['YGG_DEFAULT_COMM']
        if old_default_comm is not None:
            os.environ['YGG_DEFAULT_COMM'] = old_default_comm
        DefaultComm._reset_alias()
    return change_default_comm_w


def get_service_manager_skips(service_type, partial_commtype=None,
                              check_running=False):
    r"""Create a list of conditions and skip messages."""
    from yggdrasil.tools import is_comm_installed
    from yggdrasil.services import create_service_manager_class
    out = []
    if partial_commtype is not None:
        out.append(
            (not is_comm_installed(partial_commtype, language='python'),
             f"Communicator type '{partial_commtype}' not installed."))
    cls = create_service_manager_class(service_type=service_type)
    out.append(
        (not cls.is_installed(),
         f"Service type '{service_type}' not installed."))
    assert(not check_running)
    # if check_running and cls.is_installed():
    #     cli = IntegrationServiceManager(service_type=service_type,
    #                                     commtype=partial_commtype,
    #                                     for_request=True)
    #     out.append(
    #         (not cli.is_running,
    #          f"Service of type {service_type} not running."))
    return out


@pytest.fixture(scope="session")
def check_service_manager_settings():
    r"""Check that the requested settings are available, skipping if not."""
    def check_service_manager_settings_w(service_type, partial_commtype=None):
        skips = get_service_manager_skips(service_type,
                                          partial_commtype=partial_commtype)
        for s in skips:
            if s[0]:
                pytest.skip(s[1])
    return check_service_manager_settings_w


@pytest.fixture(scope="session")
def running_service(pytestconfig, check_service_manager_settings,
                    project_dir):
    r"""Context manager to run and clean-up an integration service."""
    manager = pytestconfig.pluginmanager
    plugin_class = manager.get_plugin('pytest_cov').CovPlugin
    with_coverage = False
    for x in manager.get_plugins():
        if isinstance(x, plugin_class):
            with_coverage = True
            break

    @contextlib.contextmanager
    def running_service_w(service_type, partial_commtype=None):
        from yggdrasil.services import (
            IntegrationServiceManager)
        if ((((service_type, partial_commtype) == ('flask', 'rmq'))
             and platform._is_win)):
            pytest.skip("excluded on windows")
        check_service_manager_settings(service_type, partial_commtype)
        model_repo = "https://github.com/cropsinsilico/yggdrasil_models_test/models"
        log_level = logging.ERROR
        args = [sys.executable, "-m", "yggdrasil", "integration-service-manager",
                f"--service-type={service_type}"]
        if partial_commtype is not None:
            args.append(f"--commtype={partial_commtype}")
        args += ["start", f"--model-repository={model_repo}",
                 f"--log-level={log_level}"]
        process_kws = {}
        if with_coverage:
            script_path = os.path.expanduser(os.path.join('~', 'run_server.py'))
            process_kws['cwd'] = project_dir
            if platform._is_win:  # pragma: windows
                process_kws['creationflags'] = subprocess.CREATE_NEW_PROCESS_GROUP
            lines = [
                'from yggdrasil.services import IntegrationServiceManager',
                'srv = IntegrationServiceManager(']
            if service_type is not None:
                lines[-1] += f'service_type=\'{service_type}\''
                if partial_commtype is not None:
                    lines[-1] += ', '
            if partial_commtype is not None:
                lines[-1] += f'commtype=\'{partial_commtype}\''
            lines[-1] += ')'
            lines += ['assert(not srv.is_running)',
                      f'srv.start_server(with_coverage={with_coverage},',
                      f'                 log_level={log_level},'
                      f'                 model_repository=\'{model_repo}\')']
            with open(script_path, 'w') as fd:
                fd.write('\n'.join(lines))
            args = [sys.executable, script_path]
        verify_flask = (service_type == 'flask')
        if verify_flask:
            # Flask is the default, verify that it is selected
            service_type = None
        cli = IntegrationServiceManager(service_type=service_type,
                                        commtype=partial_commtype,
                                        for_request=True)
        if verify_flask:
            assert(cli.service_type == 'flask')
        assert(not cli.is_running)
        p = subprocess.Popen(args, **process_kws)
        try:
            cli.wait_for_server()
            yield cli
            cli.stop_server()
            assert(not cli.is_running)
            p.wait(10)
        finally:
            if p.returncode is None:  # pragma: debug
                p.terminate()
            if with_coverage:
                if os.path.isfile(script_path):
                    os.remove(script_path)
    return running_service_w


# Utility fixtures
@pytest.fixture(scope="session")
def wait_on_function(timeout, polling_interval):
    r"""Wait on a function to time out."""
    from yggdrasil.multitasking import wait_on_function

    def wrapped_wait_on_function(*args, **kwargs):
        kwargs.setdefault("timeout", timeout)
        kwargs.setdefault("polling_interval", polling_interval)
        return wait_on_function(*args, **kwargs)
    return wrapped_wait_on_function


@pytest.fixture
def run_once(request):
    r"""Fixture indicating that the test should only be run once."""
    key = (request.cls, request.function.__name__)
    if key in _test_registry:
        pytest.skip(f"{request.cls.__name__}.{request.function.__name__} "
                    f"already ran")
    _test_registry.append(key)


@pytest.fixture
def pprint_diff():
    r"""Get the diff between the pprint.pformat string for two objects."""
    def pprint_diff_w(x, y):
        import difflib
        import pprint
        from yggdrasil import tools
        tools.print_encoded('\n'.join(difflib.ndiff(
            pprint.pformat(x).splitlines(),
            pprint.pformat(y).splitlines())))
    return pprint_diff_w


@pytest.fixture(scope="session")
def check_required_languages(pytestconfig):
    r"""Check if a set of languages is enabled/disabled."""
    def check_required_languages_w(required_languages):
        enabled = pytestconfig.getoption("--language")
        disabled = pytestconfig.getoption("--skip-language")
        if enabled and (not all(x in enabled for x in required_languages)):
            pytest.skip(f"One or more required languages "
                        f"({required_languages}) not enabled")
        if disabled and any(x in disabled for x in required_languages):
            pytest.skip(f"One or more required languages "
                        f"({required_languages}) disabled")
        for x in required_languages:
            drv = import_component('model', x)
            if not drv.is_installed():
                pytest.skip(f"{x} language not installed")
    return check_required_languages_w


@pytest.fixture(scope="session")
def check_required_comms(on_mpi):
    r"""Check if a set of comms is installed."""
    from yggdrasil.components import import_component

    def check_required_comms_w(required_comms, language="python"):
        if isinstance(required_comms, str):
            required_comms = [required_comms]
        for x in required_comms:
            comm = import_component('comm', x)
            if not comm.is_installed(language=language):
                pytest.skip(f"{x} communicator not installed for "
                            f"{language} language")
            if comm._commtype == 'mpi' and (not on_mpi):
                pytest.skip("MPI communicator requires an MPI process")
    return check_required_comms_w


@pytest.fixture(scope="class")
def recv_message_list(timeout, wait_on_function, nested_approx):
    r"""Continue receiving from a receive instance until flag is False
    (or an empty messages is received and break_on_empty is True). On
    receipt of a False flag, the recieved message is checked against the
    EOF message.

    Args:
        recv_inst (yggdrasil.communication.CommBase.CommBase): Comm
            instance that should be received from.
        expected_result (list, optional): A list of messages that the
            recieved messages should be compared against. Defaults to None
            and is ignored.
        break_on_empty (bool, optional): If True, messages will stop being
            received from the communication instance when an empty message
            is received. Defaults to False.

    Returns:
       list: Received messages.

    """
    def wrapped_recv_message_list(recv_inst, expected_result=None,
                                  break_on_empty=False):
        r"""Continue receiving until flag is False."""
        msg_list = []

        def recv_element():
            if recv_inst.is_closed:
                return True
            flag, msg_recv = recv_inst.recv(timeout)
            if flag:
                if break_on_empty and recv_inst.is_empty_recv(msg_recv):
                    return True
                msg_list.append(msg_recv)
            else:
                assert(msg_recv == recv_inst.eof_msg)
            return (not flag)
        wait_on_function(recv_element, timeout=timeout)
        if expected_result is not None:
            assert(msg_list == nested_approx(expected_result))
        return msg_list
    return wrapped_recv_message_list


@pytest.fixture(scope="session")
def assert_equal_file_contents():
    r"""Assert that the contents of two files are equivalent.

    Args:
        a (object): Contents of first file for comparison.
        b (object): Contents of second file for comparison.

    Raises:
        AssertionError: If the contents are not equal.

    """
    import difflib

    def assert_equal_file_contents_w(a, b):
        if a != b:  # pragma: debug
            odiff = '\n'.join(list(difflib.Differ().compare(a, b)))
            raise AssertionError(('File contents do not match expected '
                                  'result Diff:\n%s') % odiff)
    return assert_equal_file_contents_w


@pytest.fixture(scope="session")
def check_file_exists(wait_on_function):
    r"""Check that a file exists.

    Args:
        fname (str): Full path to the file that should be checked.
        timeout (float, optional): Time that should be waited when checking
            the file's existance. Defaults to 2.

    """
    def check_file_exists_w(fname, timeout=2):
        wait_on_function(lambda: os.path.isfile(fname), timeout=timeout,
                         on_timeout=f"File '{fname}' does not exist")
    return check_file_exists_w


@pytest.fixture(scope="session")
def check_file_size(wait_on_function):
    r"""Check that file is the correct size.

    Args:
        fname (str): Full path to the file that should be checked.
        fsize (int): Size that the file should be in bytes.
        timeout (float, optional): Time that should be waited when checking
            the file size. Defaults to 2.

    """
    def check_file_size_w(fname, fsize, timeout=2):
        result = None
        if isinstance(fsize, (bytes, str)):
            result = fsize
            fsize = len(result)

        def on_timeout():  # pragma: debug
            if (result is not None) and (fsize < 200):
                print(f"Expected:\n{result}\n"
                      f"Actual:\n{open(fname, 'r').read()}")
            raise AssertionError(f"File size ({os.stat(fname).st_size}), "
                                 f"dosn't match expected size ({fsize}).")
        wait_on_function(lambda: os.stat(fname).st_size == fsize,
                         timeout=timeout, on_timeout=on_timeout)
    return check_file_size_w


@pytest.fixture(scope="session")
def check_file_contents(assert_equal_file_contents):
    r"""Check that the contents of a file are correct.

    Args:
        fname (str): Full path to the file that should be checked.
        result (str): Contents of the file.

    """
    def check_file_contents_w(fname, result):
        ocont = open(fname, 'r').read()
        assert_equal_file_contents(ocont, result)
    return check_file_contents_w


@pytest.fixture(scope="session")
def check_file(check_file_exists, check_file_size, check_file_contents):
    r"""Check that a file exists, is the correct size, and has the correct
    contents.

    Args:
        fname (str): Full path to the file that should be checked.
        result (str): Contents of the file.

    """
    def check_file_w(fname, result):
        check_file_exists(fname)
        check_file_size(fname, len(result))
        check_file_contents(fname, result)
    return check_file_w


# Equality fixtures
@pytest.fixture(scope="session")
def pandas_equality():
    r"""Comparison operation for pandas DataFrames."""
    def pandas_equality_w(a, b):
        return a.equals(b)
    return pandas_equality_w


@pytest.fixture
def pandas_equality_patch(monkeypatch, pandas_equality):
    r"""Patch pandas DataFrame so that equals is used instead of '=='"""
    import pandas
    with monkeypatch.context() as m:
        m.setattr(pandas.DataFrame, '__eq__', pandas_equality)
        yield


@pytest.fixture(scope="session")
def unyts_equality(nested_approx):
    r"""Comparison operation for unyt quantities and arrays."""
    import unyt

    def unyts_equality_w(a, b):
        if not isinstance(b, unyt.array.unyt_array):
            return False
        if a.units != b.units:
            return False
        return a.to_ndarray() == nested_approx(b.to_ndarray())
    return unyts_equality_w


@pytest.fixture
def unyts_equality_patch(monkeypatch, unyts_equality):
    r"""Patch unyt array so that data and units considered."""
    import unyt
    with monkeypatch.context() as m:
        m.setattr(unyt.array.unyt_array, '__eq__', unyts_equality)
        yield
        

@pytest.fixture(scope="session")
def functions_equality():
    def functions_equality_w(a, b):
        a_str = f"{a.__module__}.{a.__name__}"
        b_str = f"{b.__module__}.{b.__name__}"
        if not (a_str.endswith(b_str) or b_str.endswith(a_str)):
            return False
        return a.__dict__ == b.__dict__
    return functions_equality_w


@pytest.fixture(scope="session")
def nested_approx(patch_equality, pandas_equality):
    r"""Nest pytest.approx for assertion."""
    from collections import OrderedDict
    import pandas
    import unyt

    def nested_approx_(x, **kwargs):
        if isinstance(x, dict):
            return {k: nested_approx_(v, **kwargs) for k, v in x.items()}
        elif isinstance(x, OrderedDict):
            return OrderedDict(
                [(k, nested_approx_(v, **kwargs)) for k, v in x.items()])
        elif isinstance(x, list):
            return [nested_approx_(xx, **kwargs) for xx in x]
        elif isinstance(x, tuple):
            return tuple([nested_approx_(xx, **kwargs) for xx in x])
        elif isinstance(x, pandas.DataFrame):
            return x
        elif isinstance(x, (unyt.array.unyt_quantity, unyt.array.unyt_array)):
            # from yggdrasil.units import get_ureg
            # units = str(x.units)
            # y = pytest.approx(x, **kwargs)
            # dtype = x.to_ndarray().dtype
            # return x.__class__(y, units, dtype=dtype, registry=get_ureg())
            return x
        return pytest.approx(x, **kwargs)
    return nested_approx_


@pytest.fixture(scope="session")
def patch_equality():
    def patch_equality_w(obj, method):
        class EqualityWrapper:
            def __init__(self, x):
                self.x = x

            def __str__(self):
                return f"EqualityWrapper({self.x!s})"

            def __repr__(self):
                return f"EqualityWrapper({self.x!r})"
            
            def __eq__(self, other):
                if not isinstance(other, self.x.__class__):
                    return False
                return method(self.x, other)
        return EqualityWrapper(obj)
    return patch_equality_w


# Fixtures for monitoring/managing resources
_dont_verify_count_fds = False
_dont_verify_count_comms = False
_dont_verify_count_threads = False


@pytest.fixture(scope="session", autouse=True)
def init_mp():
    r"""Initialize multiprocessing."""
    from yggdrasil.multitasking import mp_ctx_spawn
    yield mp_ctx_spawn.RLock()


@pytest.fixture(scope="session", autouse=True)
def init_zmq():
    r"""Create a socket to remove initial fd count."""
    from yggdrasil.communication.ZMQComm import _global_context
    if _global_context:
        import zmq
        s = _global_context.socket(zmq.PUSH)
        s.close()


# @pytest.fixture(autouse=True)
# def ensure_gc():
#     gc.collect()
#     yield
#     gc.collect()


@pytest.fixture
def first_test():
    r"""Stand-in for first test."""
    return True


@pytest.fixture(scope="session")
def register_weakref():
    r"""Register a weak ref for use by another fixture."""
    def register_weakref_w(x):
        import weakref
        global _weakref_registry
        _weakref_registry.append(weakref.ref(x))
    return register_weakref_w


@pytest.fixture(scope="session")
def close_comm():
    r"""Close a communicator."""
    def close_comm_w(comm):
        comm.close()
        comm.disconnect()
        assert(comm.is_closed)
        del comm
    return close_comm_w


@pytest.fixture(scope="session")
def count_comms(communicator_types):
    r"""Count the number of communicators in existence."""
    def count_comms_w(classes=None):
        from yggdrasil.communication import import_comm
        if classes is None:
            classes = communicator_types
        return sum(import_comm(k).comm_count() for k in classes)
    return count_comms_w


@pytest.fixture(scope="session")
def count_fds():
    r"""Count the number of file descriptors."""
    def count_fds_w():
        import psutil
        from yggdrasil import platform
        proc = psutil.Process()
        if platform._is_win:  # pragma: windows
            out = proc.num_handles()
        else:
            out = proc.num_fds()
        return out
    return count_fds_w


@pytest.fixture(scope="session")
def list_fds():
    r"""Get a list of file descriptors."""
    from yggdrasil import tools

    def list_fds_w():
        return tools.get_fds()
    return list_fds_w


@pytest.fixture(scope="session")
def track_fds(list_fds):
    r"""Track the creation of fds."""
    from yggdrasil.tools import track_fds as track_fds_w
    return track_fds_w


@pytest.fixture(scope="session")
def log_resource_counts(count_comms, count_fds):
    r"""Log resource counts."""
    import threading

    def log_resource_counts_w(prefix=""):
        logger.debug(f"{prefix}comms={count_comms()}, fds={count_fds()}, "
                     f"threads={threading.active_count()}")
    return log_resource_counts_w


@pytest.fixture
def disable_verify_count_threads():
    global _dont_verify_count_threads
    _dont_verify_count_threads = True
    yield


@pytest.fixture
def optionally_disable_verify_count_fds():
    def optionally_disable_verify_count_fds_w():
        global _dont_verify_count_fds
        _dont_verify_count_fds = True
    return optionally_disable_verify_count_fds_w


@pytest.fixture
def disable_verify_count_fds(optionally_disable_verify_count_fds):
    optionally_disable_verify_count_fds()
    yield


@pytest.fixture
def disable_verify_count_comms():
    global _dont_verify_count_comms
    _dont_verify_count_comms = True
    yield
    
    
@pytest.fixture
def verify_count_threads(wait_on_function):
    r"""Assert that all threads created during a test are cleaned up."""
    import threading
    global _dont_verify_count_thrads
    _dont_verify_count_thrads = False
    nthread = threading.active_count()
    yield

    if not _dont_verify_count_thrads:
        def on_timeout():  # pragma: debug
            threads = '\n\t'.join([x.name for x in threading.enumerate()])
            raise AssertionError(f"{threading.active_count()} threads "
                                 f"running, but the test started with "
                                 f"{nthread}. Running threads:\n\t{threads}")
        # Subtract one as it will be the thread checking function
        wait_on_function(lambda: threading.active_count() <= (nthread + 1),
                         on_timeout=on_timeout)


@pytest.fixture
def verify_count_comms(wait_on_function, count_comms, communicator_types):
    r"""Verify that comms created during a test are cleaned up."""
    global _dont_verify_count_comms
    _dont_verify_count_comms = False
    ncomm = count_comms()
    yield

    if not _dont_verify_count_comms:
        def on_timeout():  # pragma: debug
            comms = '\n\t'.join([f"{x}:\t{count_comms([x])}"
                                 for x in communicator_types])
            raise AssertionError(f"{count_comms()} comms "
                                 f"in registry, but the test started with "
                                 f"{ncomm}. Available comms:\n\t{comms}")
        wait_on_function(lambda: count_comms() <= ncomm,
                         on_timeout=on_timeout)


@pytest.fixture
def verify_count_fds(wait_on_function, first_test, count_fds,
                     init_zmq, init_mp):
    r"""Verify that file descriptors created during a test are cleaned up."""
    global _dont_verify_count_fds
    _dont_verify_count_fds = False
    nfds = count_fds()
    yield
    gc.collect()
    if not (first_test or _dont_verify_count_fds or platform._is_win):
        def on_timeout():  # pragma: debug
            global _weakref_registry
            for x in _weakref_registry:
                if x():
                    import pprint
                    import pdb
                    refs = gc.get_referrers(x())
                    print(f"{len(refs)} references remain")
                    pprint.pprint(refs)
                    print(f'FDS: {count_fds()}')
                    pdb.set_trace()
            raise AssertionError(f"{count_fds()} file descriptors are open, "
                                 f"but the test started with {nfds}.")
        wait_on_function(lambda: count_fds() <= nfds,
                         on_timeout=on_timeout)


@pytest.fixture
def cleanup_communicators(communicator_types):
    r"""Cleanup communicators."""
    yield
    from yggdrasil.communication import cleanup_comms
    for x in communicator_types:
        cleanup_comms(x)


# MPI utilities
_global_tag = 0
_mpi_error_exchange = None


@pytest.fixture(scope="session")
def mpi_comm():
    r"""MPI communicator."""
    try:
        from mpi4py import MPI
        return MPI.COMM_WORLD
    except ImportError:
        return None


@pytest.fixture(scope="session")
def mpi_size(mpi_comm):
    r"""int: Size of MPI job."""
    if mpi_comm is None:
        return 1
    else:
        return mpi_comm.Get_size()


@pytest.fixture(scope="session")
def mpi_rank(mpi_comm):
    r"""int: MPI rank of the current process."""
    if mpi_comm is None:
        return 0
    else:
        return mpi_comm.Get_rank()


@pytest.fixture(scope="session")
def on_mpi(mpi_size):
    r"""bool: True if this is an MPI run."""
    return (mpi_size > 1)


def new_mpi_exchange():
    from yggdrasil.multitasking import MPIErrorExchange
    global _mpi_error_exchange
    global _global_tag
    if _mpi_error_exchange is None:
        _mpi_error_exchange = MPIErrorExchange(global_tag=_global_tag)
    else:
        _global_tag = _mpi_error_exchange.global_tag
        _mpi_error_exchange.reset(global_tag=_global_tag)
    return _mpi_error_exchange


@pytest.fixture(scope="session")
def adv_global_mpi_tag():
    def adv_global_mpi_tag_w(value=1):
        global _mpi_error_exchange
        assert(_mpi_error_exchange is not None)
        out = _mpi_error_exchange.global_tag
        _mpi_error_exchange.global_tag += value
        return out
    return adv_global_mpi_tag_w


@pytest.fixture(scope="session")
def sync_mpi_exchange():
    def sync_mpi_exchange_w(*args, **kwargs):
        global _mpi_error_exchange
        assert(_mpi_error_exchange is not None)
        return _mpi_error_exchange.sync(*args, **kwargs)
    return sync_mpi_exchange_w


# Method of raising errors when other process fails
# https://docs.pytest.org/en/latest/example/simple.html#
# making-test-result-information-available-in-fixtures
@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()
    # set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"
    setattr(item, "rep_" + rep.when, rep)


@pytest.fixture(autouse=True)
def sync_mpi_result(request, on_mpi):
    r"""Synchronize results between MPI ranks."""
    mpi_exchange = None
    if on_mpi:
        mpi_exchange = new_mpi_exchange()
        mpi_exchange.sync()
    yield
    if on_mpi:
        failure = (request.node.rep_setup.failed
                   or getattr(getattr(request.node, 'rep_call', None),
                              'failed', False))
        mpi_exchange.finalize(failure)


# Monkey patch pytest-cov plugin with MPI Barriers to prevent multiple
# MPI processes from attempting to modify the .coverage data file at
# the same time and limit the coverage output to the rank 0 process
@pytest.fixture(scope="session", autouse=True)
def finalize_mpi(request, on_mpi, mpi_comm, mpi_rank, mpi_size):
    """Slow down the exit on MPI processes to prevent collision in access
    to .coverage file."""
    if not on_mpi:
        return
    manager = request.config.pluginmanager
    plugin_class = manager.get_plugin('pytest_cov').CovPlugin
    plugin = None
    for x in manager.get_plugins():
        if isinstance(x, plugin_class):
            plugin = x
            break
    if not plugin:  # pragma: no cover
        return
    old_finish = getattr(plugin.cov_controller, 'finish')

    def new_finish():
        mpi_comm.Barrier()
        for _ in range(mpi_rank):
            mpi_comm.Barrier()
        old_finish()
        # These lines come after coverage collection
        for _ in range(mpi_rank, mpi_size):  # pragma: testing
            mpi_comm.Barrier()  # pragma: testing
        mpi_comm.Barrier()  # pragma: testing

    plugin.cov_controller.finish = new_finish
    if mpi_rank != 0:

        def new_is_worker(session):  # pragma: testing
            return True

        plugin._is_worker = new_is_worker



import functools
import weakref
from yggdrasil import tools, multitasking
from yggdrasil.communication import (
    RMQComm, NoMessages, TemporaryCommunicationError)
from yggdrasil.communication.RMQComm import pika


class RMQTaskLoop(multitasking.YggTaskLoop):
    r"""Task loop for RMQ consumer."""

    def __init__(self, comm, *args, **kwargs):
        self.comm = weakref.proxy(comm)
        super(RMQTaskLoop, self).__init__(*args, **kwargs)
    
    def atexit(self):  # pragma: debug
        if self.comm.is_interface:
            if self.comm.direction == 'send' and self.comm.is_open:
                self.comm.send_eof()
                self.comm.linger()
                self.join(10.0)
            self.comm.close()
        super(RMQTaskLoop, self).atexit()


class RMQAsyncComm(RMQComm.RMQComm):
    r"""Class for handling asynchronous RabbitMQ communications. It is not
    recommended to use this class as it can leave hanging threads if not
    closed propertly. The normal RMQComm will cover most use cases.

    Args:
        name (str): The environment variable where the comm address is stored.
        dont_open (bool, optional): If True, the connection will not be opened.
            Defaults to False.
        **kwargs: Additional keyword arguments are passed to CommBase.

    Attributes:
        times_connected (int): Number of times that this connections has been
            established.
        rmq_thread (multitasking.YggTask): Thread used to run IO loop.

    """
    
    _commtype = 'rmq_async'
    _schema_subtype_description = ('Asynchronous RabbitMQ connection.')
    _disconnect_attr = (RMQComm.RMQComm._disconnect_attr
                        + ['rmq_thread', '_consuming', '_reconnecting',
                           '_buffered_messages', '_deliveries',
                           '_external_close'])

    def _init_before_open(self, **kwargs):
        r"""Initialize null variables and RMQ async thread."""
        self.original_queue = None
        self.times_connected = 0
        self.rmq_thread_count = 0
        self.rmq_thread = self.new_run_thread()
        self._consuming = multitasking.ProcessEvent(task_method='thread')
        self._reconnecting = multitasking.ProcessEvent(task_method='thread')
        self._reconnect_delay = 0
        self._prefetch_count = 2
        self._buffered_messages = multitasking.Queue(task_method='thread')
        self._deliveries = multitasking.LockedDict(task_method='thread')
        self._external_close = multitasking.Event(task_method='thread')
        # self._publish_interval = 0
        self._acked = 0
        self._nacked = 0
        self._message_number = 0
        super(RMQAsyncComm, self)._init_before_open(**kwargs)
        self._opening.stopped.add_callback(self._reconnecting.stop)

    @property
    def rmq_lock(self):
        r"""Lock associated with RMQ ioloop thread."""
        return self.rmq_thread.lock

    def new_run_thread(self, name=None):
        r"""Get a new thread for running."""
        if name is None:
            name = self.name
        self.rmq_thread_count += 1
        return RMQTaskLoop(
            self, name=f'{name}.RMQThread{self.rmq_thread_count}',
            target=self.run_thread, task_method='thread',
            daemon=self.is_interface)

    def reset_for_reconnection(self):
        r"""Reset variables in preparation for reconnection."""
        self.connection = None
        self.channel = None
        self._acked = 0
        self._nacked = 0
        self._message_number = 0
        self._consuming.started.clear()
        self._consuming.stopped.clear()
        self._closing.started.clear()
        self._closing.stopped.clear()
        self._consumer_tag = None
        if self.direction == 'send':  # pragma: intermittent
            for k in list(self._deliveries.keys()):
                self._deliveries.pop(k)

    def check_for_close(self):
        r"""bool: Check if close has been called from the main process."""
        if self._external_close.is_set():  # pragma: debug
            self._closing.start()
            if self._reconnecting.is_running():
                self._reconnecting.stop()
            self.connection.ioloop.stop()
            return True
        return False

    def run_thread(self):
        r"""Connect to the connection and begin the IO loop."""
        class BreakLoop(multitasking.BreakLoopException):
            def __init__(solf, *args, **kwargs):
                self.stop()
                super(BreakLoop, solf).__init__(*args, **kwargs)

        if self._closing.has_started():
            raise BreakLoop("closing")
        try:
            self.debug('')
            self.connect()
            self.connection.ioloop.start()
            self._consuming.stop()
            self.debug("returning")
        except BaseException as e:  # pragma: debug
            self.error("Error in RMQ thread %s: %s", type(e), e)
            raise BreakLoop(e)
        self.close_queue(skip_unbind=True)
        self.close_channel()
        self.close_connection()
        if self._closing.has_started():
            self._closing.stop()
        if self._reconnecting.is_running():
            if self._consuming.has_started():
                self._reconnect_delay = 0
            else:
                self._reconnect_delay += 1
            self._reconnect_delay = min(self._reconnect_delay, 30)
            self.info(f'Reconnecting after {self._reconnect_delay} seconds')
            self._external_close.wait(self._reconnect_delay)
            if self._external_close.is_set():  # pragma: debug
                self._reconnecting.stop()
                raise BreakLoop("external close")
            self.reset_for_reconnection()

    def start_run_thread(self):
        r"""Start the run thread and wait for it to finish."""
        with self.rmq_lock:
            if self.rmq_thread.was_started:
                return
            self._opening.start()
            self.rmq_thread.start()
        # Wait for connection to be established
        self._opening.stopped.wait(self.timeout)
        # Check that connection was established
        if not self.rmq_thread.is_alive():  # pragma: debug
            self.stop()
            raise Exception("Connection ioloop could not be established.")
        if not self.is_open:  # pragma: debug
            self.stop(call_on_thread=True)
            raise RuntimeError("Connection never finished opening ")

    def bind(self):
        r"""Declare queue to get random new queue."""
        if self._opening.has_started() or self._closing.has_started():
            return
        self.start_run_thread()  # Start ioloop in a new thread
        # Register queue
        if not self.queue:  # pragma: debug
            self.error("Queue was not initialized.")
        self.register_comm(self.address, (self.connection, self.channel))
        super(RMQComm.RMQComm, self).bind()
        multitasking.wait_on_function(
            lambda: self._opening.has_stopped() or self._closing.has_started(),
            timeout=self.timeout)
    
    def _close(self, linger=False):
        r"""Close the connection.

        Args:
            linger (bool, optional): If True, drain messages before closing the
                comm. Defaults to False.

        """
        self.stop(call_on_thread=True)
        self._closing.stopped.wait(self.timeout)
        if not self._closing.has_stopped():  # pragma: debug
            if self.connection is not None:
                self.connection.ioloop.stop()
            self._closing.stop()
            self.error("Closing has not completed, resources may be leaked")
        if not self.is_client:
            self.unregister_comm(self.address)
        with self.rmq_lock:
            self.channel = None
            self.connection = None
        super(RMQComm.RMQComm, self)._close(linger=linger)
        if self.rmq_thread.is_alive():  # pragma: debug
            self.rmq_thread.join(self.timeout)
            if self.rmq_thread.is_alive():
                raise RuntimeError(
                    self.logger.process("Thread still running.", {})[0])

    def atexit(self):  # pragma: debug
        r"""Close operations."""
        super(RMQComm.RMQComm, self).atexit()
        
    @property
    def is_open(self):
        r"""bool: True if the connection and channel are open."""
        return (super(RMQAsyncComm, self).is_open
                or self._reconnecting.is_running())

    @property
    def n_msg_recv(self):
        r"""int: Number of messages in the queue."""
        if self.is_open:
            return self._buffered_messages.qsize()
        return 0

    @property
    def n_msg_send(self):
        r"""int: Number of messages in the queue."""
        return self.n_msg_recv  # + len(self._deliveries)
        
    def _send(self, msg, **kwargs):
        r"""Send a message.

        Args:
            msg (str, bytes): Message to be sent.
            exchange (str, optional): Exchange that message should be sent
                to. Defaults to self.exchange.
            routing_key (str, optional): Key that exchange should use to route
                the message. Defaults to self.queue.
            **kwargs: Additional keyword arguments are passed to
                :method:`pika.BlockingChannel.basic_publish`.

        Returns:
            bool: Success or failure of send.

        """
        try:
            self._buffered_messages.put((msg, kwargs), block=False)
            self.connection.ioloop.add_callback_threadsafe(
                self.publish_message)
        except multitasking.queue.Full:  # pragma: debug
            raise TemporaryCommunicationError("Queue full.")
        return True

    def _recv(self):
        r"""Receive a message.

        Returns:
            tuple (bool, obj): Success or failure of receive and received
                message.

        """
        try:
            return (True, self._buffered_messages.get(block=False))
        except multitasking.queue.Empty:
            raise NoMessages("No messages in buffer.")

    def close_channel(self):
        r"""Close the channel if it exists."""
        if self.channel:
            self.close_queue(skip_unbind=True)
            self.channel.add_callback(
                self.on_channel_closeok, [pika.spec.Channel.CloseOk])
        super(RMQAsyncComm, self).close_channel()
        
    def close_connection(self, *args, **kwargs):
        r"""Close the connection."""
        if self.direction == 'recv':
            self._consuming.stop()
        if self.connection is not None:
            if self.connection.is_closing or self.connection.is_closed:
                self.debug('Connection is closing or already closed')
            else:
                super(RMQAsyncComm, self).close_connection(*args, **kwargs)
        
    # CONNECTION
    def connect(self):
        r"""Establish the connection."""
        self.times_connected += 1
        self.connection = pika.SelectConnection(
            pika.URLParameters(self.url),
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_open_error,
            on_close_callback=self.on_connection_closed)

    def on_connection_open(self, connection):
        r"""Actions that must be taken when the connection is opened.
        Add the close connection callback and open the RabbitMQ channel."""
        self.debug('Connection opened')
        self.open_channel()

    def on_connection_open_error(self, unused_connection, err):  # pragma: debug
        r"""Actions that must be taken when the connection fails to open."""
        self.debug(f'Connection open failed: {err}')
        if not self._external_close.is_set():
            self.reconnect()

    def on_connection_closed(self, connection, reason):
        r"""Actions that must be taken when the connection is closed. Set the
        channel to None. If the connection is meant to be closing, stop the
        IO loop. Otherwise, wait 5 seconds and try to reconnect."""
        with self.rmq_lock:
            self.close_queue(skip_unbind=True)
            self.channel = None
            if self._closing.has_started() or self._external_close.is_set():
                self.debug('Connection closed')
                self._closing.stop()
            else:
                self.debug(f'Connection closed, reconnecting: {reason}')
                self.reconnect()
            self.connection.ioloop.stop()

    def stop(self, call_on_thread=False):
        r"""Stop the ioloop."""
        if call_on_thread:
            if not self.rmq_thread.is_alive():
                # Ensure that shutdown is not prevented by flags
                assert(self._closing.has_stopped())
                return
            self._external_close.set()
            if self._reconnecting.is_running():  # pragma: debug
                self._reconnecting.stopped.wait(10.0)
                return
            if self.connection is not None:
                self.connection.ioloop.add_callback_threadsafe(self.stop)
            return
        if not self._closing.has_started():
            self._closing.start()
            self.debug(f'Stopping {self.direction}')
            if self.direction == 'recv':
                if self._consuming.is_running():
                    self.stop_consuming()
                    # Not called here because KeyboardInterrupt is not used
                    # to stop the loop
                    # self.connection.ioloop.start()
                elif self.connection is not None:  # pragma: debug
                    self.connection.ioloop.stop()
                    self.close_channel()
                    self.close_connection()
                    self._closing.stop()
                self.debug('Stopped')
            else:
                self.close_channel()
                self.close_connection()

    def reconnect(self):
        r"""Try to re-establish a connection and resume a new IO loop."""
        # Stop the old IOLoop, create a new connection and start a new IOLoop
        self.debug('')
        if not self.original_queue:  # pragma: debug
            self.error("Cannot reconnect to a queue with a generated name.")
            return
        self._reconnecting.started.clear()
        self._reconnecting.stopped.clear()
        self._reconnecting.start()  # stopped by re-opening
        self._opening.stopped.clear()

    # CHANNEL
    def open_channel(self):
        r"""Open a RabbitMQ channel."""
        self.debug('Creating a new channel')
        if self.check_for_close():  # pragma: debug
            return
        self.connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        r"""Actions to perform after a channel is opened. Add the channel
        close callback and setup the exchange."""
        self.debug('Channel opened')
        self.channel = channel
        self.add_on_channel_close_callback()
        if self.check_for_close():  # pragma: debug
            return
        self.setup_exchange(self.exchange)

    def add_on_channel_close_callback(self):
        """This method tells pika to call the on_channel_closed method if
        RabbitMQ unexpectedly closes the channel."""
        self.debug('Adding channel close callback')
        self.channel.add_on_close_callback(self.on_channel_closed)

    def on_channel_closed(self, channel, reason):
        r"""Actions to perform when the channel is closed. Close the
        connection."""
        self.debug(f'Channel {channel} was closed: {reason}, {type(reason)}')
        self.close_queue(skip_unbind=True)
        kwargs = {}
        if isinstance(reason, pika.exceptions.ChannelClosedByBroker):  # pragma: debug
            self._closing.start()
            self._consuming.stop()
            kwargs['reply_code'] = reason.reply_code
            kwargs['reply_text'] = reason.reply_text
        self.channel = None
        self.close_connection(**kwargs)

    # EXCHANGE
    def setup_exchange(self, exchange_name):
        r"""Setup the exchange."""
        self.debug('Declaring exchange %s', exchange_name)
        cb = functools.partial(
            self.on_exchange_declareok, userdata=exchange_name)
        self.channel.exchange_declare(
            exchange=exchange_name,
            auto_delete=True,
            callback=cb)

    def on_exchange_declareok(self, unused_frame, userdata):
        r"""Actions to perform once an exchange is succesfully declared.
        Set up the queue."""
        self.debug('Exchange declared: %s', userdata)
        if self.check_for_close():  # pragma: debug
            return
        self.setup_queue(self.queue)

    # QUEUE
    def setup_queue(self, queue_name):
        r"""Set up the message queue."""
        if self.original_queue is None:
            self.original_queue = queue_name
        passive = self.original_queue.startswith('amq.')
        self.debug(f'Declaring queue {queue_name} (passive={passive})')
        cb = functools.partial(self.on_queue_declareok, userdata=queue_name)
        self.channel.queue_declare(queue=queue_name, callback=cb,
                                   exclusive=False, passive=passive)

    def on_queue_declareok(self, method_frame, userdata):
        r"""Actions to perform once the queue is succesfully declared. Bind
        the queue."""
        queue_name = userdata
        self.debug(f'Binding {self.exchange} to {queue_name}')
        if not self.queue:
            self.address += method_frame.method.queue
        cb = functools.partial(self.on_bindok, userdata=queue_name)
        if self.check_for_close():  # pragma: debug
            return
        self.channel.queue_bind(
            queue_name,
            self.exchange,
            callback=cb)
        
    def on_bindok(self, unused_frame, userdata):
        r"""Actions to perform once the queue is succesfully bound. Start
        consuming messages."""
        self.debug('Queue bound')
        if self.check_for_close():  # pragma: debug
            return
        if self.direction == 'recv':
            self.set_qos()
        else:
            self.start_publishing()
            self.debug('Finished opening producer')

    # CONSUMER (RECV) SPECIFIC
    def set_qos(self):
        self.channel.basic_qos(
            prefetch_count=self._prefetch_count,
            callback=self.on_basic_qos_ok)

    def on_basic_qos_ok(self, unused_frame):
        r"""Actions to perform one the qos is set."""
        self.debug('QOS set to: %d', 1)
        if self.check_for_close():  # pragma: debug
            return
        self.start_consuming()

    def start_consuming(self):
        self.debug('Issuing consumer related RPC commands')
        self.add_on_cancel_callback()
        self.consumer_tag = self.channel.basic_consume(
            self.queue, self.on_message, auto_ack=True)
        self._consuming.start()
        self._opening.stop()
        self.debug('Finished opening consumer')

    def add_on_cancel_callback(self):
        self.debug('Adding consumer cancellation callback')
        self.channel.add_on_cancel_callback(self.on_consumer_cancelled)

    def on_consumer_cancelled(self, method_frame):  # pragma: debug
        self.debug('Consumer was cancelled remotely, shutting down: %r',
                   method_frame)
        self.close_channel()

    def on_channel_closeok(self, _unused_frame):
        self.debug('Channel closed.')

    def on_message(self, _unused_channel, basic_deliver, properties, body):
        r"""Buffer received messages."""
        self.debug('Received message # %s from %s: %.100s',
                   basic_deliver.delivery_tag, properties.app_id, body)
        body = tools.str2bytes(body)
        self._buffered_messages.put(body)
        # Not required because auto_ack is True
        # self.acknowledge_message(basic_deliver.delivery_tag)

    # def acknowledge_message(self, delivery_tag):
    #     self.debug('Acknowledging message %s', delivery_tag)
    #     self.channel.basic_ack(delivery_tag, True)

    def stop_consuming(self):
        if self.channel:
            self.debug('Sending a Basic.Cancel RPC command to RabbitMQ')
            cb = functools.partial(
                self.on_cancelok, userdata=self.consumer_tag)
            self.channel.basic_cancel(self.consumer_tag, cb)

    def on_cancelok(self, unused_frame, userdata):
        r"""Actions to perform after succesfully cancelling consumption. Closes
        the channel."""
        self._consuming.stop()
        self.debug(
            'RabbitMQ acknowledged the cancellation of the consumer: %s',
            userdata)
        self.close_channel()

    # PUBLISHER (SEND) SPECIFIC
    def start_publishing(self):
        r"""Enable confirmations and begin sending messages."""
        self.debug('Issuing publisher related RPC commands')
        self.enable_delivery_confirmations()
        # self.schedule_next_message()
        self._opening.stop()  # Ready for sending
    
    def enable_delivery_confirmations(self):
        r"""Turn on delivery confirmations."""
        self.debug('Issuing Confirm.Select RPC command')
        self.channel.confirm_delivery(self.on_delivery_confirmation)

    def on_delivery_confirmation(self, method_frame):
        r"""Actions performed when a sent message is confirmed."""
        confirmation_type = method_frame.method.NAME.split('.')[1].lower()
        self.debug('Received %s for delivery tag: %i', confirmation_type,
                   method_frame.method.delivery_tag)
        if confirmation_type == 'ack':
            self._acked += 1
        elif confirmation_type == 'nack':  # pragma: intermittent
            self._nacked += 1
        self._deliveries.pop(method_frame.method.delivery_tag)
        self.debug(
            'Published %i messages, %i have yet to be confirmed, '
            '%i were acked and %i were nacked', self._message_number,
            len(self._deliveries), self._acked, self._nacked)

    # def schedule_next_message(self):
    #     r"""Schedule a new send."""
    #     self.debug('Scheduling next message for %0.1f seconds',
    #                self._publish_interval)
    #     self.connection.ioloop.call_later(self._publish_interval,
    #                                       self.publish_message)

    def publish_message(self):
        r"""Publish the next message in the list."""
        try:
            msg, kwargs = self._buffered_messages.get(block=False)
        except multitasking.queue.Empty:  # pragma: intermittent
            # self.schedule_next_message()
            return
        exchange = kwargs.pop('exchange', self.exchange)
        routing_key = kwargs.pop('routing_key', self.queue)
        kwargs.setdefault('mandatory', True)
        try:
            self.channel.basic_publish(exchange, routing_key, msg, **kwargs)
        except pika.exceptions.UnroutableError:  # pragma: debug
            self.error("Failed to publish message")
            self.stop()
            return
        self._message_number += 1
        self._deliveries[self._message_number] = (msg, kwargs)
        self.debug('Published message # %i', self._message_number)
        # self.schedule_next_message()
